# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the server-side no-operation task."""

from django.core.exceptions import ValidationError

from debusine.artifacts.models import WorkRequestResults
from debusine.test.django import TestCase


class ServerNoopTests(TestCase):
    """Test the ServerNoop task."""

    def test_execute_result_true(self) -> None:
        """If result=SUCCESS, the task returns SUCCESS."""
        noop = self.playground.create_server_task(
            task_name="servernoop",
            task_data={"result": WorkRequestResults.SUCCESS},
        ).get_task()
        self.assertEqual(noop._execute(), WorkRequestResults.SUCCESS)

    def test_execute_result_false(self) -> None:
        """If result=FAILURE, the task returns FAILURE."""
        noop = self.playground.create_server_task(
            task_name="servernoop",
            task_data={"result": WorkRequestResults.FAILURE},
        ).get_task()
        self.assertEqual(noop._execute(), WorkRequestResults.FAILURE)

    def test_execute_exception(self) -> None:
        """If exception=True, the task raises an exception."""
        noop = self.playground.create_server_task(
            task_name="servernoop", task_data={"exception": True}
        ).get_task()
        self.assertRaisesRegex(
            RuntimeError, "Client requested an exception", noop._execute
        )

    def test_no_additional_properties(self) -> None:
        """Assert no additional properties."""
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(ValidationError, error_msg):
            self.playground.create_server_task(
                task_name="servernoop", task_data={"input": {}}
            )

    def test_label(self) -> None:
        """Test get_label."""
        task = self.playground.create_server_task(
            task_name="servernoop",
            task_data={"result": WorkRequestResults.SUCCESS},
        ).get_task()
        self.assertEqual(task.get_label(), "noop")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for task input fields."""

import re
from types import SimpleNamespace
from typing import Any, cast, override
from unittest import mock

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianBinaryPackage,
    DebianUpload,
)
from debusine.client.models import RelationType
from debusine.tasks import (
    BaseTask,
    BaseTaskWithExecutor,
    ExtraRepositoryMixin,
    TaskConfigError,
)
from debusine.tasks.executors import ExecutorImageCategory
from debusine.tasks.inputs import (
    DataFieldInput,
    EnvironmentInput,
    ExtraRepositoriesInput,
    MultiInputList,
    OptionalEnvironmentInput,
    OptionalSingleInput,
    SingleInputList,
    SuiteArchiveInput,
    TaskInput,
    UploadArtifactsInput,
)
from debusine.tasks.models import (
    ExtraDebusineRepository,
    ExtraExternalRepository,
    ExtraRepository,
    LookupMultiple,
)
from debusine.tasks.server import (
    ArtifactInfo,
    CollectionInfo,
    MultipleArtifactInfo,
    TaskDatabaseInterface,
)
from debusine.tasks.tests.helper_mixin import FakeTaskDatabase
from debusine.test import TestCase


class ConcreteTaskInput(TaskInput[str]):
    """Concrete task input."""

    def __init__(self, value: str) -> None:
        """Resolve input field to a string constant."""
        self.value = value

    @override
    def resolve(
        self, task: BaseTask[Any, Any], task_database: TaskDatabaseInterface
    ) -> str:
        return self.value


class InputTests(TestCase):
    """Test case that can work as a container for task input fields."""

    inputs: dict[str, TaskInput[Any]] = {}

    @override
    def setUp(self) -> None:
        super().setUp()
        # Restore the fields to unresolved state
        for name in self.inputs:
            self.__dict__.pop(name, None)
        # Task database for resolving fields
        self.task_database = FakeTaskDatabase()

    def mock_task(
        self, task_class: type[BaseTask[Any, Any]] = BaseTask[Any, Any]
    ) -> BaseTask[Any, Any]:
        """Instantiate a mock task."""
        task = mock.MagicMock(spec=task_class)
        task.data = SimpleNamespace()
        return task


class TaskInputTest(InputTests):
    """Tests for :py:class:`TaskInput`."""

    field = ConcreteTaskInput("test")

    def test_input_name(self) -> None:
        self.assertEqual(self.inputs["field"].name, "field")

    def test_registered_in_inputs(self) -> None:
        self.assertIn("field", self.inputs)

    def test_inputs_subclassing(self) -> None:
        class A:
            inputs: dict[str, TaskInput[Any]] = {}
            field1 = ConcreteTaskInput("one")

        class B(A):
            field2 = ConcreteTaskInput("two")

        self.assertIn("field1", A.inputs)
        self.assertNotIn("field2", A.inputs)
        self.assertIn("field1", B.inputs)
        self.assertIn("field2", B.inputs)

    def test_not_resolved(self) -> None:
        with self.assertRaisesRegex(
            TaskConfigError, r"Cannot access unresolved input 'field'"
        ):
            self.field  # type: ignore[arg-type]

    def test_resolve(self) -> None:
        TaskInput.resolve_inputs(
            cast(BaseTask[Any, Any], self), cast(TaskDatabaseInterface, None)
        )
        self.assertEqual(self.field, "test")  # type: ignore[arg-type]


class ConcreteDataFieldInput(DataFieldInput[str]):
    """Concrete version of DataFieldInput."""

    @override
    def resolve(
        self, task: BaseTask[Any, Any], task_database: TaskDatabaseInterface
    ) -> str:
        raise NotImplementedError()


class DataFieldInputTest(InputTests):
    """Tests for :py:class:`DataFieldInput`."""

    inputs: dict[str, TaskInput[Any]] = {}
    field_default = ConcreteDataFieldInput()
    field_test = ConcreteDataFieldInput(field="test")
    field_test_nested = ConcreteDataFieldInput(field="test.nested")

    def test_field_default(self) -> None:
        self.assertEqual(
            self.inputs["field_default"].field,  # type: ignore[attr-defined]
            "field_default",
        )

    def test_resolve_field(self) -> None:
        data = SimpleNamespace()
        data.value = 1
        data.nested = SimpleNamespace()
        data.nested.value = 2

        for field, expected in (("value", 1), ("nested.value", 2)):
            with self.subTest(field=field):
                f = ConcreteDataFieldInput(field=field)
                self.assertEqual(f._resolve_field(data), expected)

    def test_resolve_field_fails(self) -> None:
        data = SimpleNamespace()
        data.nested = SimpleNamespace()

        for field in ("value", "nested.value"):
            with (
                self.subTest(Field=field),
                self.assertRaisesRegex(
                    TaskConfigError,
                    re.escape(f"Invalid input field: {field!r}"),
                ),
            ):
                f = ConcreteDataFieldInput(field=field)
                f._resolve_field(data)

    def test_get_lookup_single(self) -> None:
        field = ConcreteDataFieldInput(field="test")
        for value in None, 1, "foo":
            with (
                self.subTest(value=value),
                mock.patch.object(
                    field,
                    "_resolve_field",
                    return_value=value,
                ),
            ):
                self.assertEqual(field._get_lookup_single(mock.Mock()), value)

    def test_get_lookup_single_wrong_type(self) -> None:
        field = ConcreteDataFieldInput(field="test")
        value: Any
        for value in 3.14, [], {}:
            with (
                self.subTest(value=value),
                mock.patch.object(
                    field,
                    "_resolve_field",
                    return_value=value,
                ),
                self.assertRaises(AssertionError),
            ):
                field._get_lookup_single(mock.Mock())

    def test_get_lookup_multiple(self) -> None:
        field = ConcreteDataFieldInput(field="test")
        for value in (
            LookupMultiple((1,)),
            LookupMultiple((1, 2, "three")),
        ):
            with (
                self.subTest(value=repr(value)),
                mock.patch.object(
                    field,
                    "_resolve_field",
                    return_value=value,
                ),
            ):
                self.assertEqual(field._get_lookup_multiple(mock.Mock()), value)

    def test_get_lookup_multiple_wrong_type(self) -> None:
        field = ConcreteDataFieldInput(field="test")
        value: Any
        for value in None, 1, "foo", 3.14, [], {}:
            with (
                self.subTest(value=value),
                mock.patch.object(
                    field,
                    "_resolve_field",
                    return_value=value,
                ),
                self.assertRaises(AssertionError),
            ):
                field._get_lookup_multiple(mock.Mock())

    def test_get_lookup_list_None(self) -> None:
        field = ConcreteDataFieldInput(field="test")
        with mock.patch.object(
            field,
            "_resolve_field",
            return_value=None,
        ):
            self.assertEqual(field._get_lookup_list(mock.Mock()), [])

    def test_get_lookup_list(self) -> None:
        field = ConcreteDataFieldInput(field="test")
        for value in [], [1], [LookupMultiple((1,))]:
            with (
                self.subTest(value=repr(value)),
                mock.patch.object(
                    field,
                    "_resolve_field",
                    return_value=value,
                ),
            ):
                self.assertEqual(field._get_lookup_list(mock.Mock()), value)

    def test_get_lookup_list_wrong_type(self) -> None:
        field = ConcreteDataFieldInput(field="test")
        value: Any
        for value in 1, "foo", 3.14, {}:
            with (
                self.subTest(value=value),
                mock.patch.object(
                    field,
                    "_resolve_field",
                    return_value=value,
                ),
                self.assertRaises(AssertionError),
            ):
                field._get_lookup_list(mock.Mock())


class SingleInputListTests(InputTests):
    """Tests for :py:class:`SingleInputList`."""

    def test_empty_categories(self) -> None:
        field = SingleInputList()
        task = self.mock_task()
        with (
            mock.patch.object(
                field, "_get_lookup_list", return_value=[1, "two"]
            ),
            mock.patch.object(
                self.task_database,
                "lookup_single_artifact",
                side_effect=["foo", "bar"],
            ) as lookup_single_artifacts,
            mock.patch.object(field, "check_category") as check_category,
        ):
            self.assertEqual(
                field.resolve(task, self.task_database),
                ["foo", "bar"],
            )

        lookup_single_artifacts.assert_has_calls(
            [mock.call(1), mock.call("two")]
        )
        check_category.assert_has_calls([mock.call("foo"), mock.call("bar")])


class OptionalSingleInputTests(InputTests):
    """Tests for :py:class:`OptioanlSingleInput`."""

    def setUp(self) -> None:
        super().setUp()
        self.field = OptionalSingleInput()
        self.task = self.mock_task()
        self.enterContext(
            mock.patch.object(
                self.field, "_get_lookup_single", return_value="LOOKUP"
            )
        )
        self.check_category = self.enterContext(
            mock.patch.object(self.field, "check_category")
        )

    def resolve(self, result: str | None) -> str | None:
        with mock.patch.object(
            self.task_database,
            "lookup_single_artifact",
            return_value=result,
        ) as lookup_single_artifact:
            resolved = cast(
                str | None, self.field.resolve(self.task, self.task_database)
            )
        lookup_single_artifact.assert_called_with("LOOKUP")
        return resolved

    def test_none(self) -> None:
        self.assertEqual(self.resolve(None), None)
        self.check_category.assert_not_called()

    def test_has_value(self) -> None:
        self.assertEqual(self.resolve("foo"), "foo")
        self.check_category.assert_called_with("foo")


class MultiInputListTests(InputTests):
    """Tests for :py:class:`MultiInputList`."""

    def setUp(self) -> None:
        super().setUp()
        self.task = self.mock_task()
        self.check_category = self.enterContext(
            mock.patch("debusine.tasks.inputs.ArtifactInput.check_category")
        )

    def test_empty_categories(self) -> None:
        self.field = MultiInputList()
        with (
            mock.patch.object(
                self.field,
                "_get_lookup_list",
                return_value=[LookupMultiple((1,))],
            ),
            mock.patch.object(
                self.task_database,
                "lookup_multiple_artifacts",
                return_value="foo",
            ),
        ):
            self.assertEqual(
                self.field.resolve(self.task, self.task_database),
                ["foo"],
            )

        self.check_category.assert_not_called()

    def test_with_categories(self) -> None:
        self.field = MultiInputList(
            field="test", categories=[ArtifactCategory.BINARY_PACKAGE]
        )
        with (
            mock.patch.object(
                self.field,
                "_get_lookup_list",
                return_value=[LookupMultiple((1,)), LookupMultiple((2,))],
            ),
            mock.patch.object(
                self.task_database,
                "lookup_multiple_artifacts",
                side_effect=[["foo"], ["bar", "baz"]],
            ),
        ):
            self.assertEqual(
                self.field.resolve(self.task, self.task_database),
                [["foo"], ["bar", "baz"]],
            )

        self.check_category.assert_has_calls(
            [
                mock.call("foo", "test[0][0]"),
                mock.call("bar", "test[1][0]"),
                mock.call("baz", "test[1][1]"),
            ]
        )


class UploadArtifactsInputTests(InputTests):
    """Tests for :py:class:`UploadArtifactsInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_database = FakeTaskDatabase()

    def resolve(
        self,
        value: list[ArtifactInfo],
        categories: list[ArtifactCategory] | None = None,
    ) -> MultipleArtifactInfo:
        field = UploadArtifactsInput(
            field="test",
            categories=categories or [ArtifactCategory.BINARY_PACKAGE],
        )
        task = mock.MagicMock()
        task.data = SimpleNamespace()
        lookup = mock.MagicMock(spec=LookupMultiple)
        with (
            mock.patch.object(field, "_resolve_field", return_value=lookup),
            mock.patch.object(
                self.task_database,
                "lookup_multiple_artifacts",
                return_value=MultipleArtifactInfo(tuple(value)),
            ),
        ):
            return field.resolve(task, self.task_database)

    def test_empty(self) -> None:
        self.assertEqual(self.resolve([]), MultipleArtifactInfo())

    def test_no_uploads(self) -> None:
        binary = ArtifactInfo(
            id=1,
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="unsigned",
                srcpkg_version="1.0",
                deb_fields={
                    "Package": "hello",
                    "Version": "1.0-1",
                    "Architecture": "amd64",
                },
                deb_control_files=[],
            ),
        )
        self.assertEqual(
            self.resolve([binary]), MultipleArtifactInfo((binary,))
        )

    def test_expand_uploads(self) -> None:
        binary1 = ArtifactInfo(
            id=1,
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="unsigned",
                srcpkg_version="1.0",
                deb_fields={
                    "Package": "hello",
                    "Version": "1.0-1",
                    "Architecture": "amd64",
                },
                deb_control_files=[],
            ),
        )
        binary2 = ArtifactInfo(
            id=2,
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name="unsigned-doc",
                srcpkg_version="1.0",
                deb_fields={
                    "Package": "unsigned",
                    "Version": "1.0-1",
                    "Architecture": "all",
                },
                deb_control_files=[],
            ),
        )
        upload = ArtifactInfo(
            id=3,
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Architecture": "all",
                    "Files": [{"name": "unsigned-doc_1.0_all.deb"}],
                },
            ),
        )
        self.task_database.relations = {
            (
                (upload.id,),
                ArtifactCategory.BINARY_PACKAGE,
                RelationType.EXTENDS,
            ): [binary2]
        }
        self.assertEqual(
            [x.id for x in self.resolve([binary1, upload])],
            [x.id for x in (binary1, binary2)],
        )


class EnvironmentInputTests(InputTests):
    """Tests for :py:class:`EnvironmentInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_database = FakeTaskDatabase()
        self.task = mock.MagicMock(spec=BaseTaskWithExecutor)
        self.task.name = "TASKNAME"
        self.task.build_architecture = mock.Mock(return_value="BUILDARCH")
        self.task.backend = "BACKEND"
        self.task.data = SimpleNamespace()

    def assertResolves(self, field: EnvironmentInput) -> None:
        with mock.patch.object(field, "_resolve_field", return_value="LOOKUP"):
            self.assertEqual(
                field.resolve(
                    cast(BaseTask[Any, Any], self.task), self.task_database
                ),
                "foo",
            )

    def test_defaults(self) -> None:
        field = EnvironmentInput()
        with (
            mock.patch.object(
                self.task_database, "get_environment", return_value="foo"
            ) as get_environment,
            mock.patch.object(field, "check_category") as check_category,
        ):
            self.assertResolves(field)

        check_category.assert_called()
        get_environment.assert_called_once_with(
            "LOOKUP",
            architecture="BUILDARCH",
            backend="BACKEND",
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=None,
            try_variant="TASKNAME",
        )

    def test_set_backend_false(self) -> None:
        field = EnvironmentInput(set_backend=False)
        with (
            mock.patch.object(
                self.task_database, "get_environment", return_value="foo"
            ) as get_environment,
            mock.patch.object(field, "check_category") as check_category,
        ):
            self.assertResolves(field)

        check_category.assert_called()
        get_environment.assert_called_once_with(
            "LOOKUP",
            architecture="BUILDARCH",
            backend=None,
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=None,
            try_variant="TASKNAME",
        )

    def test_try_variant_failse(self) -> None:
        field = EnvironmentInput(try_variant=False)
        with (
            mock.patch.object(
                self.task_database, "get_environment", return_value="foo"
            ) as get_environment,
            mock.patch.object(field, "check_category") as check_category,
        ):
            self.assertResolves(field)

        check_category.assert_called()
        get_environment.assert_called_once_with(
            "LOOKUP",
            architecture="BUILDARCH",
            backend="BACKEND",
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=None,
            try_variant=None,
        )

    def test_image_category(self) -> None:
        field = EnvironmentInput(image_category=ExecutorImageCategory.TARBALL)
        with (
            mock.patch.object(
                self.task_database, "get_environment", return_value="foo"
            ) as get_environment,
            mock.patch.object(field, "check_category") as check_category,
        ):
            self.assertResolves(field)

        check_category.assert_called()
        get_environment.assert_called_once_with(
            "LOOKUP",
            architecture="BUILDARCH",
            backend="BACKEND",
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=ExecutorImageCategory.TARBALL,
            try_variant="TASKNAME",
        )

    def test_field_is_required(self) -> None:
        field = EnvironmentInput(field="test")
        self.task.data.test = None
        with self.assertRaises(AssertionError):
            (
                field.resolve(
                    cast(BaseTask[Any, Any], self.task), self.task_database
                ),
            )


class OptionalEnvironmentInputTests(InputTests):
    """Tests for :py:class:`OptionalEnvironmentInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_database = FakeTaskDatabase()
        self.task = mock.MagicMock(spec=BaseTaskWithExecutor)
        self.task.name = "TASKNAME"
        self.task.build_architecture = mock.Mock(return_value="BUILDARCH")
        self.task.backend = "BACKEND"
        self.task.data = {}

    def test_defaults(self) -> None:
        field = OptionalEnvironmentInput()
        with (
            mock.patch.object(field, "_resolve_field", return_value="LOOKUP"),
            mock.patch.object(
                self.task_database, "get_environment", return_value="foo"
            ) as get_environment,
            mock.patch.object(field, "check_category") as check_category,
        ):
            self.assertEqual(
                field.resolve(
                    cast(BaseTask[Any, Any], self.task), self.task_database
                ),
                "foo",
            )

        check_category.assert_called()
        get_environment.assert_called_once_with(
            "LOOKUP",
            architecture="BUILDARCH",
            backend="BACKEND",
            default_category=CollectionCategory.ENVIRONMENTS,
            image_category=None,
            try_variant="TASKNAME",
        )

    def test_missing(self) -> None:
        field = OptionalEnvironmentInput()
        with (
            mock.patch.object(field, "_resolve_field", return_value=None),
            mock.patch.object(
                self.task_database, "get_environment", return_value="foo"
            ) as get_environment,
            mock.patch.object(field, "check_category") as check_category,
        ):
            self.assertIsNone(
                field.resolve(
                    cast(BaseTask[Any, Any], self.task), self.task_database
                )
            )

        check_category.assert_not_called()
        get_environment.assert_not_called()


class SuiteArchiveInputTests(InputTests):
    """Tests for :py:class:`SuiteArchiveInput`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_database = FakeTaskDatabase()
        self.suite = CollectionInfo(
            id=1,
            scope_name="scope",
            workspace_name="workspace",
            category=CollectionCategory.SUITE,
            name="sid",
            data={},
        )
        self.archive = CollectionInfo(
            id=2,
            scope_name="scope",
            workspace_name="workspace",
            category=CollectionCategory.ARCHIVE,
            name="_",
            data={},
        )

    def resolve(
        self, suite: CollectionInfo, archive: CollectionInfo | None = None
    ) -> tuple[CollectionInfo, CollectionInfo | None]:
        field = SuiteArchiveInput(field="test")
        task = mock.MagicMock()
        task.data = SimpleNamespace()
        lookup = "LOOKUP"
        with (
            mock.patch.object(field, "_resolve_field", return_value=lookup),
            mock.patch.object(
                self.task_database,
                "lookup_single_collection",
                return_value=suite,
            ) as lookup_single_collection,
            mock.patch.object(
                self.task_database,
                "lookup_singleton_collection",
                return_value=archive,
            ) as lookup_singleton_collection,
        ):
            res = field.resolve(task, self.task_database)

        lookup_single_collection.assert_called_once_with(lookup)
        lookup_singleton_collection.assert_called_once_with(
            CollectionCategory.ARCHIVE
        )

        return res

    def test_suite_only(self) -> None:
        self.assertEqual(self.resolve(self.suite, None), (self.suite, None))

    def test_suite_archive(self) -> None:
        self.assertEqual(
            self.resolve(self.suite, self.archive), (self.suite, self.archive)
        )

    def test_wrong_suite_category(self) -> None:
        self.suite.category = CollectionCategory.TEST
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"^test: unexpected collection category: "
            rf"'{CollectionCategory.TEST}'\. "
            rf"Expected: '{CollectionCategory.SUITE}'$",
        ):
            self.resolve(self.suite, self.archive), (self.suite, self.archive)


class ExtraRepositoriesInputTests(InputTests):
    """Tests for :py:class:`ExtraRepositoriesInput`."""

    def setUp(self) -> None:
        super().setUp()
        self.suite_trixie_local = CollectionInfo(
            id=1,
            scope_name="scope",
            workspace_name="workspace",
            category=CollectionCategory.SUITE,
            name="trixie-local",
            data={"components": ["main"]},
        )
        self.signing_key = "\n".join(
            (
                "-----BEGIN PGP PUBLIC KEY BLOCK-----",
                "",
                "ABCDEFGHI",
                "-----END PGP PUBLIC KEY BLOCK-----",
            )
        )
        self.task_database.single_collection_lookups[
            ("trixie-local@debian:suite", None)
        ] = self.suite_trixie_local
        self.task_database.suite_signing_keys[1] = self.signing_key
        self.task_database.settings["DEBUSINE_DEBIAN_ARCHIVE_PRIMARY_FQDN"] = (
            "deb.example.com"
        )
        self.task = self.mock_task(ExtraRepositoryMixin)  # type: ignore[type-abstract]

    def resolve(
        self, value: list[ExtraRepository] | None
    ) -> list[ExtraExternalRepository] | None:
        field = ExtraRepositoriesInput(field="test")
        with mock.patch.object(field, "_resolve_field", return_value=value):
            return field.resolve(self.task, self.task_database)

    def test_empty(self) -> None:
        self.assertEqual(self.resolve(None), None)
        self.assertEqual(self.resolve([]), [])

    def test_external(self) -> None:
        repo = ExtraExternalRepository(
            url="https://example.org", suite="sid", components=["main"]
        )
        self.assertEqual(self.resolve([repo]), [repo])

    def test_debusine(self) -> None:
        resolved = self.resolve(
            [
                ExtraDebusineRepository(
                    suite="trixie-local@debian:suite", components=["main"]
                )
            ]
        )
        self.assertEqual(
            resolved,
            [
                ExtraExternalRepository(
                    url="http://deb.example.com/scope/workspace",
                    suite="trixie-local",
                    components=["main"],
                    signing_key=self.signing_key,
                )
            ],
        )

    def test_missing_components(self) -> None:
        self.suite_trixie_local.data = {}

        with self.assertRaisesRegex(
            TaskConfigError,
            "'components' not set for trixie-local@debian:suite",
        ):
            self.resolve(
                [
                    ExtraDebusineRepository(
                        suite="trixie-local@debian:suite", components=None
                    )
                ]
            )

    def test_default_components(self) -> None:
        self.suite_trixie_local.data = {
            "components": ["main", "contrib", "non-free"]
        }

        resolved = self.resolve(
            [
                ExtraDebusineRepository(
                    suite="trixie-local@debian:suite", components=None
                )
            ]
        )

        self.assertEqual(
            resolved,
            [
                ExtraExternalRepository(
                    url="http://deb.example.com/scope/workspace",
                    suite="trixie-local",
                    components=["main", "contrib", "non-free"],
                    signing_key=self.signing_key,
                )
            ],
        )

    def test_explicit_components(self) -> None:
        self.suite_trixie_local.data = {
            "components": ["main", "contrib", "non-free"]
        }

        resolved = self.resolve(
            [
                ExtraDebusineRepository(
                    suite="trixie-local@debian:suite",
                    components=["main", "contrib"],
                )
            ]
        )

        self.assertEqual(
            resolved,
            [
                ExtraExternalRepository(
                    url="http://deb.example.com/scope/workspace",
                    suite="trixie-local",
                    components=["main", "contrib"],
                    signing_key=self.signing_key,
                )
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the ExternalDebsign task."""

import debusine.worker.tags as wtags
from debusine.artifacts.models import ArtifactCategory, DebianUpload
from debusine.server.tasks.wait.models import ExternalDebsignDynamicData
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.tasks.server import ArtifactInfo
from debusine.tasks.tests.helper_mixin import FakeTaskDatabase
from debusine.test.django import TestCase


class ExternalDebsignTaskTests(TestCase):
    """Test the ExternalDebsign task."""

    def test_compute_system_required_tags(self) -> None:
        task = self.playground.create_wait_task(
            task_name="externaldebsign",
            task_data={"unsigned": 1},
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        ).get_task()
        self.assertCountEqual(
            task.compute_system_required_tags(),
            [wtags.WORKER_TYPE_NOT_ASSIGNABLE],
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        unsigned_lookup = "internal@collections/unsigned-upload"
        task_db = FakeTaskDatabase(
            single_lookups={
                # unsigned
                (unsigned_lookup, None): ArtifactInfo(
                    1,
                    ArtifactCategory.UPLOAD,
                    DebianUpload(
                        type="dpkg",
                        changes_fields={
                            "Architecture": "source",
                            "Files": [{"name": "hello_1.0.dsc"}],
                        },
                    ),
                ),
            }
        )

        task = self.playground.create_wait_task(
            task_name="externaldebsign",
            task_data={"unsigned": unsigned_lookup},
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        ).get_task()
        self.assertEqual(
            task.compute_dynamic_data(task_db),
            ExternalDebsignDynamicData(unsigned_id=1),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        task = self.playground.create_wait_task(
            task_name="externaldebsign",
            task_data={"unsigned": 1},
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        ).get_task()
        self.assertEqual(task.get_input_artifacts_ids(), [])

        task.dynamic_data = ExternalDebsignDynamicData(unsigned_id=1)
        self.assertEqual(task.get_input_artifacts_ids(), [1])

    def test_execute(self) -> None:
        """Executing the task does nothing, successfully."""
        work_request = self.playground.create_wait_task(
            task_name="externaldebsign",
            task_data={"unsigned": 1},
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        work_request.dynamic_task_data = {"unsigned_id": 1}
        work_request.save()
        task = work_request.get_task()
        self.assertTrue(task.execute())

    def test_label(self) -> None:
        """Test get_label."""
        work_request = self.playground.create_wait_task(
            task_name="externaldebsign",
            task_data={"unsigned": "internal@collections/unsigned-upload"},
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        work_request.dynamic_task_data = {"unsigned_id": 1}
        work_request.save()
        task = work_request.get_task()
        self.assertEqual(
            task.get_label(),
            "wait for external debsign for "
            "internal@collections/unsigned-upload",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the system_information."""

import subprocess
from unittest import mock
from unittest.mock import MagicMock

import psutil

import debusine.worker.tags as wtags
from debusine.tasks.models import WorkerType
from debusine.test import TestCase
from debusine.worker import system_information


class TestSystemInformation(TestCase):
    """Test for system_information.py functions."""

    def setUp(self) -> None:
        """Ensure that native_architecture() is not cached."""
        super().setUp()
        system_information.native_architecture.cache_clear()
        self.addCleanup(system_information.native_architecture.cache_clear)

    def test_total_physical_memory(self) -> None:
        """Test total_physical_memory function."""
        total_physical_memory = system_information.total_physical_memory()

        # Assert that total_physical_memory "looks good": it is an integer
        # and at least 128 MB of memory
        self.assertIsInstance(total_physical_memory, int)
        self.assertGreaterEqual(total_physical_memory, 128 * 1024 * 1024)

        # Check that it matches the underlying implementation
        self.assertEqual(total_physical_memory, psutil.virtual_memory().total)

    def test_total_cpu_count(self) -> None:
        """Test cpu_count function."""
        cpu_count = system_information.cpu_count()

        # Assert that cpu_count "looks good": it is an integer
        # and at least 1 CPU
        self.assertIsInstance(cpu_count, int)
        self.assertGreaterEqual(cpu_count, 1)

        # Check that it matches the underlying implementation
        self.assertEqual(cpu_count, psutil.cpu_count())

    def test_total_cpu_count_undefined(self) -> None:
        """Test cpu_count function when undefined."""
        with mock.patch("psutil.cpu_count", return_value=None):
            self.assertEqual(system_information.cpu_count(), 1)

    def test_native_architecture(self) -> None:
        """Test native_architecture() function."""
        architecture = system_information.native_architecture()

        # Assert that architecture "looks good": type, length, no end of line
        assert isinstance(architecture, str)
        self.assertGreaterEqual(len(architecture), 1)
        self.assertEqual(architecture, architecture.strip())

        self.assertEqual(
            architecture,
            subprocess.check_output(
                ["dpkg", "--print-architecture"], text=True
            ).strip(),
        )

    def patch_subprocess_check_output(self) -> MagicMock:
        """Patch subprocess.check_output. Return its mock."""
        patcher = mock.patch("subprocess.check_output", autospec=True)
        mocked = patcher.start()
        self.addCleanup(patcher.stop)
        return mocked

    def test_native_architecture_dpkg_not_available(self) -> None:
        """
        Test native_architecture function: return None.

        dpkg is not available or returns an error.
        """
        mocked = self.patch_subprocess_check_output()
        mocked.side_effect = FileNotFoundError()
        self.assertIsNone(system_information.native_architecture())

        system_information.native_architecture.cache_clear()
        mocked.side_effect = subprocess.CalledProcessError(
            1, ["dpkg", "--print-architecture"]
        )
        self.assertIsNone(system_information.native_architecture())

    def test_kernel_supported_architectures_arch_test_fails(self) -> None:
        """
        Test kernel_supported_architectures return None.

        arch-test returns an error.
        """
        mocked = self.patch_subprocess_check_output()

        mocked.side_effect = subprocess.CalledProcessError(
            1, ["arch-test", "-n"]
        )
        self.assertIsNone(system_information.kernel_supported_architectures())

    def test_kernel_supported_architectures(self) -> None:
        """Test kernel_supported_architectures returns a list of archs."""
        mocked = self.patch_subprocess_check_output()

        architectures = ["x1", "x2"]
        mocked.side_effect = ["x2\nx1\n"]

        self.assertEqual(
            system_information.kernel_supported_architectures(), architectures
        )

    def test_system_metadata_without_native_kernel_supported_architectures(
        self,
    ) -> None:
        """
        system_metadata() does not include architectures related keys.

        Functions native_architecture and kernel_supported_architectures
        returned None.
        """
        with (
            mock.patch(
                "debusine.worker.system_information.native_architecture",
                return_value=None,
            ) as architecture_mocked,
            mock.patch(
                "debusine.worker.system_information"
                ".kernel_supported_architectures",
                return_value=None,
            ) as kernel_supported_architectures_mocked,
        ):
            metadata = system_information.system_metadata(WorkerType.EXTERNAL)

        architecture_mocked.assert_called()
        kernel_supported_architectures_mocked.assert_called()

        self.assertNotIn("system:native_architecture", metadata)
        self.assertNotIn("system:architectures", metadata)

    def test_system_metadata_architectures_same_as_native_architecture(
        self,
    ) -> None:
        """
        system_metadata() returns valid metadata for an external worker.

        Method system_information.kernel_supported_architectures() returns
        None: architectures is the same as native_architecture.
        """
        total_physical_memory = system_information.total_physical_memory()
        cpu_count = system_information.cpu_count()
        native_architecture = system_information.native_architecture()

        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=None,
        ):
            self.assertEqual(
                system_information.system_metadata(WorkerType.EXTERNAL),
                {
                    "system:cpu_count": cpu_count,
                    "system:total_physical_memory": total_physical_memory,
                    "system:worker_type": WorkerType.EXTERNAL,
                    "system:native_architecture": native_architecture,
                    "system:architectures": [native_architecture],
                },
            )

    def test_system_metadata_architectures_different_to_native_architecture(
        self,
    ) -> None:
        """system_metadata() populates architectures."""
        native_architecture = system_information.native_architecture()
        kernel_supported_architectures = ["x1", "x2"]
        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=kernel_supported_architectures,
        ):
            self.assertDictContainsAll(
                system_information.system_metadata(WorkerType.EXTERNAL),
                {
                    "system:native_architecture": native_architecture,
                    "system:architectures": kernel_supported_architectures,
                },
            )

    def test_system_provided_tags_without_native_kernel_supported_architectures(
        self,
    ) -> None:
        """
        system_provided_tags() does not include build-arch tags.

        Functions native_architecture and kernel_supported_architectures
        returned None.
        """
        with (
            mock.patch(
                "debusine.worker.system_information.native_architecture",
                return_value=None,
            ) as architecture_mocked,
            mock.patch(
                "debusine.worker.system_information"
                ".kernel_supported_architectures",
                return_value=None,
            ) as kernel_supported_architectures_mocked,
        ):
            tags = system_information.system_provided_tags(WorkerType.EXTERNAL)

        architecture_mocked.assert_called()
        kernel_supported_architectures_mocked.assert_called()

        self.assertEqual(
            [t for t in tags if t.startswith(wtags.BUILD_ARCH_PREFIX)], []
        )

    def test_system_provided_tags_arch_same_as_native(
        self,
    ) -> None:
        """
        system_provided_tags() returns native architecture.

        Method system_information.kernel_supported_architectures() returns
        None: architectures is the same as native_architecture.
        """
        native_architecture = system_information.native_architecture()
        assert native_architecture is not None

        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=None,
        ):
            tags = system_information.system_provided_tags(WorkerType.EXTERNAL)

        self.assertEqual(
            [t for t in tags if t.startswith(wtags.BUILD_ARCH_PREFIX)],
            [wtags.BUILD_ARCH_PREFIX + native_architecture],
        )

    def test_system_provided_tags_arch_different_than_native(
        self,
    ) -> None:
        """system_provided_tags() returns all available architectures."""
        native_architecture = system_information.native_architecture()
        assert native_architecture is not None
        kernel_supported_architectures = ["x1", "x2"]
        with mock.patch(
            "debusine.worker.system_information.kernel_supported_architectures",
            return_value=kernel_supported_architectures,
        ):
            tags = system_information.system_provided_tags(WorkerType.EXTERNAL)

        self.assertCountEqual(
            [t for t in tags if t.startswith(wtags.BUILD_ARCH_PREFIX)],
            [
                wtags.BUILD_ARCH_PREFIX + native_architecture,
                wtags.BUILD_ARCH_PREFIX + "x1",
                wtags.BUILD_ARCH_PREFIX + "x2",
            ],
        )

    def test_system_metadata_celery(self) -> None:
        """system_metadata() returns valid metadata for a Celery worker."""
        total_physical_memory = system_information.total_physical_memory()
        cpu_count = system_information.cpu_count()
        native_architecture = system_information.native_architecture()

        system_metadata = system_information.system_metadata(WorkerType.CELERY)

        # system:architectures is tested in other unit tests
        system_metadata.pop("system:architectures")

        self.assertEqual(
            system_metadata,
            {
                "system:cpu_count": cpu_count,
                "system:total_physical_memory": total_physical_memory,
                "system:worker_type": WorkerType.CELERY,
                "system:native_architecture": native_architecture,
            },
        )

    def test_system_metadata_signing(self) -> None:
        """system_metadata() returns valid metadata for a signing worker."""
        total_physical_memory = system_information.total_physical_memory()
        cpu_count = system_information.cpu_count()
        native_architecture = system_information.native_architecture()

        system_metadata = system_information.system_metadata(WorkerType.SIGNING)

        # system:architectures is tested in other unit tests
        system_metadata.pop("system:architectures")

        self.assertEqual(
            system_metadata,
            {
                "system:cpu_count": cpu_count,
                "system:total_physical_memory": total_physical_memory,
                "system:worker_type": WorkerType.SIGNING,
                "system:native_architecture": native_architecture,
            },
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Base support for wait tasks."""

from abc import ABCMeta

from debusine.artifacts.models import TaskTypes
from debusine.server.tasks.base import BaseServerTask
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData


class BaseWaitTask[TD: BaseTaskData, DTD: BaseDynamicTaskData](
    BaseServerTask[TD, DTD], metaclass=ABCMeta
):
    """Base class for wait tasks."""

    TASK_TYPE = TaskTypes.WAIT



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the blhc task support on the worker."""

import itertools
from pathlib import Path
from typing import Any
from unittest.mock import call, patch

from debian.debian_support import Version

import debusine.worker.tags as wtags
from debusine.artifacts import BlhcArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianPackageBuildLog,
    WorkRequestResults,
)
from debusine.client.models import LookupResultType, LookupSingleResponse
from debusine.tasks import Blhc, TaskConfigError
from debusine.tasks.models import BlhcDynamicData
from debusine.tasks.server import ArtifactInfo
from debusine.tasks.tests.helper_mixin import (
    ExternalTaskHelperMixin,
    MockTaskDatabase,
)
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_remote_artifact,
    create_system_tarball_data,
)


class BlhcTests(ExternalTaskHelperMixin[Blhc], TestCase):
    """Test the Blhc class."""

    SAMPLE_TASK_DATA = {
        "input": {"artifact": 421},
        "build_architecture": "amd64",
        "environment": "debian/match:codename=bookworm",
    }

    build_log = DebianPackageBuildLog(
        source="hello",
        version="1.0",
        architecture="amd64",
        filename="hello_1.0_amd64.buildlog",
    )

    def setUp(self) -> None:
        super().setUp()
        self.configure_task()

    def tearDown(self) -> None:
        """Delete directory to avoid ResourceWarning with python -m unittest."""
        if self.task._debug_log_files_directory is not None:
            self.task._debug_log_files_directory.cleanup()
        super().tearDown()

    def configure_task(
        self,
        task_data: dict[str, Any] | None = None,
        override: dict[str, Any] | None = None,
        remove: list[str] | None = None,
    ) -> None:
        """Perform further setup."""
        super().configure_task(task_data, override, remove)

        self.task_db = MockTaskDatabase(
            environment=ArtifactInfo(
                id=1,
                category=ArtifactCategory.SYSTEM_TARBALL,
                data=create_system_tarball_data(),
            ),
            input_artifact=ArtifactInfo(
                id=421,
                category=ArtifactCategory.PACKAGE_BUILD_LOG,
                data=self.build_log,
            ),
        )

    def test_configure_fails_with_missing_required_data(
        self,
    ) -> None:
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"input": {}})

    def test_configure_with_extra_flag(self) -> None:
        """Configuration included "extra_flags". Saved, no exceptions."""
        extra_flags = ["--bindnow"]

        directory = self.create_temporary_directory()
        (directory / "file.build").write_text("")

        self.configure_task(override={"extra_flags": extra_flags})

        self.patch_prepare_executor_instance()

        self.assertEqual(self.task.data.extra_flags, extra_flags)

        self.assertTrue(self.task.configure_for_execution(directory))

    def test_configure_with_invalid_flag(self) -> None:
        """Configuration included "extra_flags". Saved, no exceptions."""
        extra_flags = ["--unknown-flag"]

        directory = self.create_temporary_directory()
        (directory / "file.build").write_text("")

        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"extra_flags": extra_flags})

    def test_configure_additional_properties_in_input(self) -> None:
        """Assert no additional properties in task_data input."""
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(TaskConfigError, error_msg):
            self.configure_task(
                task_data={
                    "input": {
                        "artifact": 5,
                        "some_additional_property": 87,
                    }
                }
            )

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:blhc:version:1",
            ],
        )

    def test_compute_system_required_tags(self) -> None:
        self.assertCountEqual(
            self.task.compute_system_required_tags(),
            [wtags.EXECUTOR_PREFIX + "unshare"],
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BlhcDynamicData(
                environment_id=1,
                input_artifact_id=421,
                input_artifact_data=self.build_log,
                subject="hello",
                parameter_summary="hello_1.0 (amd64)",
                runtime_context=None,
            ),
        )

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            Blhc.inputs["input_artifact"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.PACKAGE_BUILD_LOG,),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = BlhcDynamicData(
            input_artifact_id=1, environment_id=2
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2])

    def test_configure_for_execution_from_artifact_error_no_build_logs(
        self,
    ) -> None:
        """configure_for_execution() no .build files: return False."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        (file1 := download_directory / "file1.dsc").write_text("")
        (file2 := download_directory / "file2.changes").write_text("")

        self.assertFalse(self.task.configure_for_execution(download_directory))

        files = sorted([str(file1), str(file2)])
        assert self.task._debug_log_files_directory
        log_file_contents = (
            Path(self.task._debug_log_files_directory.name)
            / "configure_for_execution.log"
        ).read_text()
        self.assertEqual(
            log_file_contents,
            f"No *.build files to be analyzed. Files: {files}\n",
        )

    def test_blhc_supports_debian_option_stretch(self) -> None:
        with patch.object(
            self.task,
            "get_package_version",
            return_value=Version("0.07+20161116+gitbf41976-1"),
        ):
            self.assertFalse(self.task._blhc_supports_debian_option())

    def test_blhc_supports_debian_option_buster(self) -> None:
        with patch.object(
            self.task, "get_package_version", return_value=Version("0.09-1")
        ):
            self.assertTrue(self.task._blhc_supports_debian_option())

    def test_cmdline_without_debian_option(self) -> None:
        self.task._blhc_target = self.create_temporary_file()

        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=False,
        ) as mock_blhc_supports_debian_option:
            cmdline = self.task._cmdline()

        self.assertEqual(cmdline, ["blhc", str(self.task._blhc_target)])
        mock_blhc_supports_debian_option.assert_called_once_with()

    def test_cmdline_with_debian_option(self) -> None:
        self.task._blhc_target = self.create_temporary_file()

        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=True,
        ) as mock_blhc_supports_debian_option:
            cmdline = self.task._cmdline()

        self.assertEqual(
            cmdline, ["blhc", "--debian", str(self.task._blhc_target)]
        )
        mock_blhc_supports_debian_option.assert_called_once_with()

    def test_cmdline_with_extra_flags(self) -> None:
        """Cmdline add extra_flags."""
        extra_flags = ["--bindnow"]
        self.configure_task(override={"extra_flags": extra_flags})
        self.task._blhc_target = self.create_temporary_file()

        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=True,
        ):
            cmdline = self.task._cmdline()

        self.assertIn(extra_flags[0], cmdline)
        self.assertEqual(cmdline[-1], str(self.task._blhc_target))

    def test_task_result_empty_file_success(self) -> None:
        """task_result() for an empty file return SUCCESS."""
        directory = self.create_temporary_directory()
        (directory / Blhc.CAPTURE_OUTPUT_FILENAME).write_text("")

        self.configure_task()
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_failure_exit_code(self) -> None:
        """task_result() for an exit code of 8 return FAILURE."""
        directory = self.create_temporary_directory()
        (directory / Blhc.CAPTURE_OUTPUT_FILENAME).write_text("")

        self.configure_task()
        self.assertEqual(
            self.task.task_result(8, directory), WorkRequestResults.FAILURE
        )

    def test_upload_artifacts(self) -> None:
        """upload_artifact() and relation_create() is called."""
        exec_dir = self.create_temporary_directory()

        # Create file that will be attached when uploading the artifacts
        blhc_output = exec_dir / self.task.CAPTURE_OUTPUT_FILENAME
        blhc_output.write_text("NONVERBOSE BUILD:    Compiling gkrust\n")

        # Used to verify the relations
        self.task._source_artifacts_ids = [1]
        self.task.dynamic_data = BlhcDynamicData(
            input_artifact_id=2,
            input_artifact_data=self.build_log,
            environment_id=1,
        )

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        workspace_name = "testing"

        uploaded_artifacts = [
            create_remote_artifact(id=10, workspace=workspace_name),
        ]

        debusine_mock.upload_artifact.side_effect = uploaded_artifacts

        # self.task.workspace_name is set by the Worker
        # and is the workspace that downloads the artifact
        # containing the files needed for Blhc
        self.task.workspace_name = workspace_name

        # The worker set self.task.work_request_id of the task
        work_request_id = 147
        self.task.work_request_id = work_request_id

        self.task.upload_artifacts(
            exec_dir, execution_result=WorkRequestResults.SUCCESS
        )

        # Debusine Mock upload_artifact expected calls
        upload_artifact_calls = []
        blhc_artifact = BlhcArtifact.create(
            blhc_output=blhc_output, build_log=self.build_log
        )

        upload_artifact_calls.append(
            call(
                blhc_artifact,
                workspace=workspace_name,
                work_request=work_request_id,
            )
        )

        # Debusine mock relation_create expected calls
        relation_create_calls = []
        for uploaded_artifact, source_artifact_id in itertools.product(
            uploaded_artifacts, self.task._source_artifacts_ids
        ):
            relation_create_calls.append(
                call(uploaded_artifact.id, source_artifact_id, "relates-to")
            )

        # Assert that the artifacts were uploaded and relations created
        debusine_mock.upload_artifact.assert_has_calls(upload_artifact_calls)
        debusine_mock.relation_create.assert_has_calls(relation_create_calls)

    def test_execute(self) -> None:
        """Test full (mocked) execution."""
        build_log = DebianPackageBuildLog(
            source="hello",
            version="1.0",
            architecture="i386",
            filename="hello_1.0_i386.buildlog",
        )
        self.configure_task(override={"input": {'artifact': 1}})
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = BlhcDynamicData(
            input_artifact_id=1, input_artifact_data=build_log, environment_id=2
        )
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponse(
            result_type=LookupResultType.ARTIFACT, artifact=1
        )
        debusine_mock.download_artifact.return_value = True
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=2, workspace=self.task.workspace_name
        )

        self.assertTrue(self.task.fetch_input(download_directory))

        debusine_mock.download_artifact.assert_called_once_with(
            1, download_directory, tarball=False, workspace="testing"
        )

        f1_contents = "Compiling gkrust v0.1.0 (/build/rust)"
        (file1 := download_directory / "file.build").write_text(f1_contents)

        self.patch_prepare_executor_instance()

        self.assertTrue(self.task.configure_for_execution(download_directory))

        self.assertEqual(self.task._blhc_target, file1)
        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=True,
        ):
            self.assertEqual(
                self.task._cmdline(),
                [
                    "blhc",
                    "--debian",
                    str(file1),
                ],
            )

        # Create file that will be attached when uploading the artifacts
        blhc_output = download_directory / self.task.CAPTURE_OUTPUT_FILENAME
        blhc_output.write_text(
            "NONVERBOSE BUILD:    Compiling gkrust v0.1.0 (/build/rust)\n"
        )

        self.task.upload_artifacts(
            download_directory, execution_result=WorkRequestResults.SUCCESS
        )

        debusine_mock.upload_artifact.assert_called_once_with(
            BlhcArtifact.create(
                blhc_output=download_directory / Blhc.CAPTURE_OUTPUT_FILENAME,
                build_log=build_log,
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )
        debusine_mock.relation_create.assert_called_once_with(
            2, 1, "relates-to"
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the __main__ (Worker's CLI launcher)."""

import contextlib
import io
import signal
from unittest import mock
from unittest.mock import AsyncMock

from debusine.test import TestCase
from debusine.worker.__main__ import main


class MainTests(TestCase):
    """Tests for __main__ (Worker's CLI launcher)."""

    DEFAULT_ARGV = ['--log-file', '/dev/null']

    def setUp(self) -> None:
        """Initialize test."""
        super().setUp()

        self.default_sigint_handler = signal.getsignal(signal.SIGINT)
        self.default_sigterm_handler = signal.getsignal(signal.SIGTERM)

        self.patch_config_handler()

    def tearDown(self) -> None:
        """Restore default signals (__main__.main() changes them)."""
        signal.signal(signal.SIGINT, self.default_sigint_handler)
        signal.signal(signal.SIGTERM, self.default_sigterm_handler)

        super().tearDown()

    def test_worker_version(self) -> None:
        stdout = io.StringIO()

        with (
            contextlib.redirect_stdout(stdout),
            self.assertRaisesSystemExit(0),
            mock.patch(
                "debusine.worker.__main__.debusine_version",
                return_value="0.1.0",
            ),
        ):
            main(["--version"])

        self.assertEqual("debusine-worker 0.1.0\n", stdout.getvalue())

    def test_worker_version_uninstalled(self) -> None:
        stdout = io.StringIO()

        with (
            contextlib.redirect_stdout(stdout),
            self.assertRaisesSystemExit(0),
            mock.patch(
                "debusine.worker.__main__.debusine_version", return_value=None
            ),
        ):
            main(["--version"])

        self.assertEqual(
            "debusine-worker uninstalled-version\n", stdout.getvalue()
        )

    def patch_config_handler(self) -> None:
        """Mock ConfigHandler to avoid trying to access config.ini."""
        config_handler_patcher = mock.patch(
            'debusine.worker._worker.ConfigHandler', autospec=True
        )
        self.config_handler_mocked = config_handler_patcher.start()
        self.config_handler_mocked().log_level = None
        self.addCleanup(config_handler_patcher.stop)

    def patch_worker_main(self) -> mock.MagicMock:
        """Patch and return mock for Worker.main()."""
        patch = mock.patch(
            "debusine.worker.__main__.Worker.main", autospec=True
        )
        main_mocked = patch.start()
        self.addCleanup(patch.stop)

        return main_mocked

    def test_main_entry_point(self) -> None:
        """Worker's main entry point calls Worker().main()."""
        mocked_main = self.patch_worker_main()

        main(self.DEFAULT_ARGV)

        mocked_main.assert_awaited_once()
        self.config_handler_mocked.assert_called()

    def patch_worker(self) -> mock.MagicMock:
        """Patch Worker class and returns its mock."""
        patch = mock.patch("debusine.worker.__main__.Worker")
        worker_mocked = patch.start()
        self.addCleanup(patch.stop)

        worker_mocked().main = AsyncMock()

        return worker_mocked

    def test_main_entry_point_log_file(self) -> None:
        """Worker main entry point calls run_worker() with the right params."""
        worker_mocked = self.patch_worker()

        main(['--log-file', '/dev/null'])

        worker_mocked.assert_called_with(log_file='/dev/null', log_level=None)

    def test_main_entry_point_no_log_file(self) -> None:
        """Worker main entry point calls run_worker() with the right params."""
        worker_mocked = self.patch_worker()

        main([])

        worker_mocked.assert_called_with(log_file=None, log_level=None)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the lookup views."""

from typing import Any, ClassVar

from django.urls import reverse
from rest_framework import status

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.models import CollectionItem, WorkRequest, Worker, Workspace
from debusine.server.views.lookups import LookupMultipleView, LookupSingleView
from debusine.server.views.rest import IsWorkerAuthenticated
from debusine.tasks.models import LookupMultiple
from debusine.test.django import TestCase, TestResponseType


class LookupSingleViewTests(TestCase):
    """Tests for LookupSingleView."""

    workspace: ClassVar[Workspace]
    worker: ClassVar[Worker]
    work_request: ClassVar[WorkRequest]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        cls.workspace = cls.playground.create_workspace(
            name="public", public=True
        )
        cls.worker = Worker.objects.create_with_fqdn(
            "worker-test", token=cls.playground.create_bare_token()
        )
        cls.work_request = cls.playground.create_worker_task(
            worker=cls.worker,
            task_name="noop",
            workspace=cls.workspace,
        )

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        self.assertIn(
            IsWorkerAuthenticated, LookupSingleView.permission_classes
        )

        response = self.client.post(
            reverse("api:lookup-single"),
            data={
                "lookup": "",
                "work_request": self.work_request.id,
                "expect_type": LookupChildType.ARTIFACT,
            },
            content_type="application/json",
        )

        self.assertEqual(
            response.json(),
            {
                "title": "Error",
                "detail": "Authentication credentials were not provided.",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_work_request_not_found(self) -> None:
        """The work request must exist."""
        assert self.worker.token is not None
        response = self.client.post(
            reverse("api:lookup-single"),
            data={
                "lookup": "x",
                "work_request": self.work_request.id + 1,
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.worker.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response, "Error", detail_pattern="Invalid pk"
        )

    def test_unauthorized(self) -> None:
        """The work request must be assigned to this worker."""
        assert self.worker.token is not None
        another_worker = Worker.objects.create_with_fqdn(
            "another-worker-test", token=self.playground.create_bare_token()
        )
        another_work_request = self.playground.create_worker_task(
            worker=another_worker, task_name="noop"
        )

        response = self.client.post(
            reverse("api:lookup-single"),
            data={
                "lookup": "x",
                "work_request": another_work_request.id,
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.worker.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            f"Work request {another_work_request.id} is not assigned to the "
            f"authenticated worker",
            status_code=status.HTTP_401_UNAUTHORIZED,
        )

    def post_lookup(
        self,
        lookup: int | str,
        expect_type: LookupChildType,
        default_category: CollectionCategory | None = None,
    ) -> TestResponseType:
        """Make a lookup request."""
        data = {
            "lookup": lookup,
            "work_request": self.work_request.id,
            "expect_type": expect_type,
            "default_category": default_category,
        }
        assert self.worker.token is not None
        return self.client.post(
            reverse("api:lookup-single"),
            data=data,
            headers={"token": self.worker.token.key},
            content_type="application/json",
        )

    def test_success_string(self) -> None:
        """A string lookup succeeds and returns an artifact ID."""
        artifact = self.playground.create_artifact(workspace=self.workspace)[0]

        response = self.post_lookup(
            f"{artifact.id}@artifacts", LookupChildType.ARTIFACT
        )

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "result_type": LookupResultType.ARTIFACT,
                "collection_item": None,
                "artifact": artifact.id,
                "collection": None,
            },
        )

    def test_success_integer(self) -> None:
        """An integer lookup succeeds and returns an artifact ID."""
        artifact = self.playground.create_artifact(workspace=self.workspace)[0]

        response = self.post_lookup(artifact.id, LookupChildType.ARTIFACT)

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "result_type": LookupResultType.ARTIFACT,
                "collection_item": None,
                "artifact": artifact.id,
                "collection": None,
            },
        )

    def test_key_error(self) -> None:
        """The view returns an error if the lookup returns no items."""
        response = self.post_lookup(
            "nonsense@artifacts", LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "No matches",
            detail_pattern="'nonsense@artifacts' does not exist or is hidden",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_lookup_error(self) -> None:
        """The view returns an error if the lookup is invalid."""
        response = self.post_lookup(
            "internal@collections", LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "internal@collections is only valid in the context of a "
                "workflow"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_wrong_workspace(self) -> None:
        """The item must be in a visible workspace."""
        workspace = self.playground.create_workspace(name="test")
        artifact = self.playground.create_artifact(workspace=workspace)[0]
        lookup = f"{artifact.id}@artifacts"

        response = self.post_lookup(lookup, LookupChildType.ARTIFACT)

        self.assertResponseProblem(
            response,
            "No matches",
            detail_pattern=f"{lookup!r} does not exist or is hidden",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_wrong_type(self) -> None:
        """The item must be of the correct type."""
        artifact = self.playground.create_artifact(workspace=self.workspace)[0]
        lookup = f"{artifact.id}@artifacts"

        response = self.post_lookup(lookup, LookupChildType.COLLECTION)

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=f"{lookup!r} is of type 'artifact'"
            " instead of expected 'collection'",
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_default_category(self) -> None:
        """The client can request a default category."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )

        response = self.post_lookup("collection", LookupChildType.COLLECTION)

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "'collection' does not specify a category and the context "
                "does not supply a default"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

        response = self.post_lookup(
            "collection",
            LookupChildType.COLLECTION,
            default_category=CollectionCategory.TEST,
        )

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "result_type": LookupResultType.COLLECTION,
                "collection_item": None,
                "artifact": None,
                "collection": collection.id,
            },
        )


class LookupMultipleViewTests(TestCase):
    """Tests for LookupMultipleView."""

    workspace: ClassVar[Workspace]
    worker: ClassVar[Worker]
    work_request: ClassVar[WorkRequest]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        cls.workspace = cls.playground.create_workspace(
            name="public", public=True
        )
        cls.worker = Worker.objects.create_with_fqdn(
            "worker-test", token=cls.playground.create_bare_token()
        )
        cls.work_request = cls.playground.create_worker_task(
            worker=cls.worker, task_name="noop", workspace=cls.workspace
        )

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        self.assertIn(
            IsWorkerAuthenticated, LookupMultipleView.permission_classes
        )

        response = self.client.post(
            reverse("api:lookup-multiple"),
            data={
                "lookup": {"collection": "test"},
                "work_request": self.work_request.id,
                "expect_type": LookupChildType.ARTIFACT,
            },
            content_type="application/json",
        )

        self.assertEqual(
            response.json(),
            {
                "title": "Error",
                "detail": "Authentication credentials were not provided.",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_work_request_not_found(self) -> None:
        """The work request must exist."""
        assert self.worker.token is not None
        response = self.client.post(
            reverse("api:lookup-multiple"),
            data={
                "lookup": {"collection": "test"},
                "work_request": self.work_request.id + 1,
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.worker.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response, "Error", detail_pattern="Invalid pk"
        )

    def test_unauthorized(self) -> None:
        """The work request must be assigned to this worker."""
        assert self.worker.token is not None
        another_worker = Worker.objects.create_with_fqdn(
            "another-worker-test", token=self.playground.create_bare_token()
        )
        another_work_request = self.playground.create_worker_task(
            worker=another_worker, task_name="noop"
        )

        response = self.client.post(
            reverse("api:lookup-multiple"),
            data={
                "lookup": {"collection": "test"},
                "work_request": another_work_request.id,
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.worker.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            f"Work request {another_work_request.id} is not assigned to the "
            f"authenticated worker",
            status_code=status.HTTP_401_UNAUTHORIZED,
        )

    def post_lookup(
        self,
        lookup: dict[str, Any] | list[str | dict[str, Any]],
        expect_type: LookupChildType,
        default_category: CollectionCategory | None = None,
    ) -> TestResponseType:
        """Make a lookup request."""
        data = {
            "lookup": lookup,
            "work_request": self.work_request.id,
            "expect_type": expect_type,
            "default_category": default_category,
        }
        assert self.worker.token
        return self.client.post(
            reverse("api:lookup-multiple"),
            data=data,
            headers={"token": self.worker.token.key},
            content_type="application/json",
        )

    def test_success(self) -> None:
        """The lookup succeeds and returns an artifact ID."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        artifacts = [
            self.playground.create_artifact(category=ArtifactCategory.TEST)[0]
            for _ in range(2)
        ]
        collection_items = [
            CollectionItem.objects.create(
                parent_collection=collection,
                name=f"artifact{i}",
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.TEST,
                artifact=artifact,
                data={},
                created_by_user=self.playground.get_default_user(),
            )
            for i, artifact in enumerate(artifacts)
        ]

        response = self.post_lookup(
            {"collection": f"collection@{CollectionCategory.TEST}"},
            LookupChildType.ARTIFACT,
        )

        data = self.assertAPIResponseOk(response)
        self.assertCountEqual(
            data,
            [
                {
                    "result_type": LookupResultType.ARTIFACT,
                    "collection_item": collection_item.id,
                    "artifact": artifact.id,
                    "collection": None,
                }
                for artifact, collection_item in zip(
                    artifacts, collection_items
                )
            ],
        )

    def test_normalized(self) -> None:
        """The view accepts fully-normalized lookups."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        artifacts = [
            self.playground.create_artifact(category=ArtifactCategory.TEST)[0]
            for _ in range(2)
        ]
        collection_items = [
            CollectionItem.objects.create(
                parent_collection=collection,
                name=f"artifact{i}",
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.TEST,
                artifact=artifact,
                data={},
                created_by_user=self.playground.get_default_user(),
            )
            for i, artifact in enumerate(artifacts)
        ]

        lookup = LookupMultiple.model_validate(
            {"collection": f"collection@{CollectionCategory.TEST}"}
        ).model_dump()
        self.assertEqual(
            lookup,
            (
                {
                    "category": None,
                    "child_type": "artifact",
                    "collection": f"collection@{CollectionCategory.TEST}",
                    "data_matchers": (),
                    "name_matcher": None,
                    "lookup_filters": (),
                },
            ),
        )

        response = self.post_lookup(lookup, LookupChildType.ARTIFACT)

        data = self.assertAPIResponseOk(response)
        self.assertCountEqual(
            data,
            [
                {
                    "result_type": LookupResultType.ARTIFACT,
                    "collection_item": collection_item.id,
                    "artifact": artifact.id,
                    "collection": None,
                }
                for artifact, collection_item in zip(
                    artifacts, collection_items
                )
            ],
        )

    def test_deserialization_error(self) -> None:
        """The view returns an error if it cannot deserialize the lookup."""
        response = self.post_lookup(
            "nonsense",  # type: ignore[arg-type]
            LookupChildType.ARTIFACT,
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.headers["Content-Type"], "application/problem+json"
        )
        self.assertEqual(response.json()["title"], "Cannot deserialize lookup")
        self.assertRegex(
            response.json()["validation_errors"]["lookup"],
            r"^1 validation error for LookupMultiple",
        )

    def test_key_error(self) -> None:
        """The view returns an empty list if the lookup returns no items."""
        response = self.post_lookup(
            {"collection": "nonexistent@test"}, LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "One of the lookups returned no matches",
            detail_pattern="'nonexistent@test' does not exist or is hidden",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_lookup_error(self) -> None:
        """The view returns an error if the lookup is invalid."""
        response = self.post_lookup(
            {"collection": "internal@collections"},
            LookupChildType.ARTIFACT,
        )

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "internal@collections is only valid in the context of a "
                "workflow"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_wrong_workspace(self) -> None:
        """The item must be in a visible workspace."""
        workspace = self.playground.create_workspace(name="test")
        self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=workspace
        )

        response = self.post_lookup(
            [{"collection": f"collection@{CollectionCategory.TEST}"}],
            LookupChildType.ARTIFACT,
        )

        self.assertResponseProblem(
            response,
            "One of the lookups returned no matches",
            detail_pattern=(
                f"'collection@{CollectionCategory.TEST}' does not exist or is "
                f"hidden"
            ),
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_wrong_type(self) -> None:
        """The lookup must be of the correct type."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        self.playground.create_bare_data_item(collection, "bare")
        lookup = {
            "collection": f"collection@{CollectionCategory.TEST}",
            "child_type": "bare",
        }

        response = self.post_lookup(lookup, LookupChildType.COLLECTION)

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                f"Result from {lookup!r} is of type 'bare' instead of expected"
                f" 'collection'"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_default_category(self) -> None:
        """The client can request a default category."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        artifact = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )[0]
        collection_item = CollectionItem.objects.create(
            parent_collection=collection,
            name="artifact",
            child_type=CollectionItem.Types.ARTIFACT,
            category=ArtifactCategory.TEST,
            artifact=artifact,
            data={},
            created_by_user=self.playground.get_default_user(),
        )

        response = self.post_lookup(
            {"collection": "collection"}, LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "'collection' does not specify a category and the context "
                "does not supply a default"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

        response = self.post_lookup(
            {"collection": "collection"},
            LookupChildType.ARTIFACT,
            default_category=CollectionCategory.TEST,
        )

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            [
                {
                    "result_type": LookupResultType.ARTIFACT,
                    "collection_item": collection_item.id,
                    "artifact": artifact.id,
                    "collection": None,
                }
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Workflow to update environments."""

import copy
from dataclasses import dataclass
from typing import Any, cast, override

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
)
from debusine.server.workflows.base import Workflow
from debusine.server.workflows.models import (
    UpdateEnvironmentsWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BaseDynamicTaskData,
    BaseTaskData,
    MmDebstrapData,
    SystemImageBuildData,
)
from debusine.tasks.server import TaskDatabaseInterface


@dataclass(frozen=True)
class ActionVariables:
    """Variables to pass to an update-collection-with-artifacts action."""

    codename: str
    variant: str | None
    backend: str | None

    def to_dict(self) -> dict[str, str]:
        """Make a dict for an update-collection-with-artifacts action."""
        variables = {"codename": self.codename}
        if self.variant is not None:
            variables["variant"] = self.variant
        if self.backend is not None:
            variables["backend"] = self.backend
        return variables


@dataclass(frozen=True)
class PlannedEnvironment:
    """Relevant data to build environments with a specific task."""

    task_data_template: dict[str, Any]
    architecture: str
    codenames: list[str]
    backends: list[str | None]
    variants: list[str | None]


class UpdateEnvironmentsWorkflow(
    Workflow[UpdateEnvironmentsWorkflowData, BaseDynamicTaskData]
):
    """Build tarballs/images and add them to an environments collection."""

    TASK_NAME = "update_environments"

    def _make_action_variables(
        self,
        *,
        codenames: list[str],
        variants: list[str | None],
        backends: list[str | None],
    ) -> list[ActionVariables]:
        """
        Return collection update action variables for a child work request.

        Each item in the returned list should be used to build an event
        reaction that adds artifacts to a collection.
        """
        action_variables: list[ActionVariables] = []
        variants_or_none: list[str | None] = list(variants)
        if not variants_or_none:
            variants_or_none.append(None)
        backends_or_none: list[str | None] = list(backends)
        if not backends_or_none:
            backends_or_none.append(None)
        for codename in codenames:
            for variant in variants_or_none:
                for backend in backends_or_none:
                    action_variables.append(
                        ActionVariables(
                            codename=codename, variant=variant, backend=backend
                        )
                    )
        return action_variables

    def _add_child(
        self,
        task_name: str,
        task_data_class: type[BaseTaskData],
        template: dict[str, Any],
        *,
        category: ArtifactCategory,
        display_name_prefix: str,
        vendor: str,
        codenames: list[str],
        architecture: str,
        variants: list[str | None],
        backends: list[str | None],
    ) -> None:
        """Create a child work request."""
        # Prepare variables and compute string representations
        codename = codenames[0]
        backend_string = (
            "|".join(b or "no-backend" for b in backends) or "no-backend"
        )
        variant_string = (
            "|".join(v or "no-variant" for v in variants) or "no-variant"
        )
        step = (
            f"{task_name}/{codename}/{architecture}/"
            f"{backend_string}/{variant_string}"
        )
        display_name = (
            f"{display_name_prefix} for {codename} / "
            f"{architecture} / {backend_string} / {variant_string}"
        )

        task_data = copy.deepcopy(template)
        task_data.setdefault("bootstrap_options", {})["architecture"] = (
            architecture
        )
        for repository in task_data.get("bootstrap_repositories", []):
            repository.setdefault("suite", codename)

        # Create the work request with its event reactions
        wr = self.work_request_ensure_child_worker(
            task_name=task_name,
            task_data=task_data_class(**task_data),
            workflow_data=WorkRequestWorkflowData(
                step=step, display_name=display_name
            ),
        )
        for variables in self._make_action_variables(
            codenames=codenames, variants=variants, backends=backends
        ):
            wr.add_event_reaction(
                "on_success",
                ActionUpdateCollectionWithArtifacts(
                    collection=(f"{vendor}@{CollectionCategory.ENVIRONMENTS}"),
                    variables=variables.to_dict(),
                    artifact_filters={"category": category},
                ),
            )

    @override
    def build_dynamic_data(
        self,
        task_database: TaskDatabaseInterface,  # noqa: U100
    ) -> BaseDynamicTaskData:
        """
        Compute dynamic data for this workflow.

        :subject: ``vendor``
        :parameter_summary: ``vendor`` being updated and number of targets
        """
        tarballs, images = self._plan_work_requests()
        count = len(tarballs) + len(images)
        summary = f"{self.data.vendor} ({count} environments)"

        return BaseDynamicTaskData(
            subject=self.data.vendor, parameter_summary=summary
        )

    def _plan_work_requests(
        self,
    ) -> tuple[list[PlannedEnvironment], list[PlannedEnvironment]]:
        images = []
        tarballs = []
        for target in self.data.targets:
            codenames = (
                target.codenames
                if isinstance(target.codenames, list)
                else [target.codenames]
            )
            variants: list[str | None] = (
                target.variants
                if isinstance(target.variants, list)
                else [target.variants]
            )
            backends: list[str | None] = (
                cast(list[str | None], target.backends)
                if isinstance(target.backends, list)
                else [target.backends]
            )

            for codename in codenames:
                codename_aliases = [codename] + target.codename_aliases.get(
                    codename, []
                )

                for architecture in target.architectures:
                    if target.mmdebstrap_template is not None:
                        environment = PlannedEnvironment(
                            task_data_template=target.mmdebstrap_template,
                            architecture=architecture,
                            codenames=codename_aliases,
                            backends=backends,
                            variants=variants,
                        )
                        tarballs.append(environment)

                    if target.simplesystemimagebuild_template is not None:
                        environment = PlannedEnvironment(
                            task_data_template=target.simplesystemimagebuild_template,
                            architecture=architecture,
                            codenames=codename_aliases,
                            backends=backends,
                            variants=variants,
                        )
                        images.append(environment)

        return tarballs, images

    @override
    def populate(self) -> None:
        """Create child work requests for all targets and architectures."""
        tarballs, images = self._plan_work_requests()

        for tarball in tarballs:
            self._add_child(
                "mmdebstrap",
                MmDebstrapData,
                tarball.task_data_template,
                category=ArtifactCategory.SYSTEM_TARBALL,
                display_name_prefix="Build tarball",
                vendor=self.data.vendor,
                codenames=tarball.codenames,
                architecture=tarball.architecture,
                variants=tarball.variants,
                backends=tarball.backends,
            )

        for image in images:
            self._add_child(
                "simplesystemimagebuild",
                SystemImageBuildData,
                image.task_data_template,
                category=ArtifactCategory.SYSTEM_IMAGE,
                display_name_prefix="Build image",
                vendor=self.data.vendor,
                codenames=image.codenames,
                architecture=image.architecture,
                variants=image.variants,
                backends=image.backends,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for scheduler tags backend functions."""

from unittest import TestCase

from debusine.server.tags import Tagset
from debusine.utils.tag_utils import DerivationRules, ProvenanceRestrictions


class TagsetTests(TestCase):
    """Tests for :py:class:`Tagset`."""

    def test_init(self) -> None:
        t = Tagset()
        self.assertEqual(t.restrictions, [])
        self.assertEqual(t.derivation_rules, [])
        self.assertEqual(t.tags, set())
        self.assertEqual(t.log, [])

    def test_plain(self) -> None:
        t = Tagset()
        t.add("test", None)
        t.add("test", ["foo"])
        t.add("test", {"bar", "baz"})
        t.finalize()
        self.assertEqual(t.tags, {"foo", "bar", "baz"})

    def test_restrictions(self) -> None:
        restrictions = ProvenanceRestrictions("test")
        restrictions.add_exact("group", ["system"])
        restrictions.add_exact("test", ["system"])
        t = Tagset()
        t.restrictions.append(restrictions)
        t.add("system", {"test", "foo"})
        t.add("workspace", {"group", "bar"})
        t.finalize()
        self.assertEqual(t.tags, {"test", "foo", "bar"})
        self.assertEqual(
            t.log, ["test provenance restrictions forbids tags group"]
        )

    def test_derivation(self) -> None:
        derivation = DerivationRules(provenance="system")
        derivation.add_rule("dm or dd", {"debian"})
        t = Tagset()
        t.derivation_rules.append(derivation)
        t.add("system", {"dm", "foo"})
        t.finalize()
        self.assertEqual(t.tags, {"dm", "debian", "foo"})
        self.assertEqual(t.log, ["system derivation rules adds tags debian"])

    def test_derivation_exists_but_does_not_match(self) -> None:
        derivation = DerivationRules(provenance="system")
        derivation.add_rule("dm or dd", {"debian"})
        t = Tagset()
        t.derivation_rules.append(derivation)
        t.add("system", {"foo"})
        t.finalize()
        self.assertEqual(t.tags, {"foo"})
        self.assertEqual(t.log, [])

    def test_derivation_restrictions(self) -> None:
        restrictions = ProvenanceRestrictions("test")
        restrictions.add_exact("group", ["system"])
        derivation = DerivationRules(provenance="workspace")
        derivation.add_rule("dm or dd", {"debian"})
        derivation.add_rule("foo", {"group"})
        t = Tagset()
        t.restrictions.append(restrictions)
        t.derivation_rules.append(derivation)
        t.add("system", {"test", "foo", "dm"})
        t.finalize()
        self.assertEqual(t.tags, {"test", "foo", "dm", "debian"})
        self.assertEqual(
            t.log,
            [
                "workspace derivation rules adds tags debian, group",
                "test provenance restrictions forbids tags group",
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the package_publish workflow."""

import logging
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, ClassVar

import debusine.worker.tags as wtags
from debusine.artifacts.models import CollectionCategory, TaskTypes
from debusine.db.models import (
    Artifact,
    Collection,
    TaskDatabase,
    WorkRequest,
    Workspace,
)
from debusine.db.playground import scenarios
from debusine.server.celery import run_server_task
from debusine.server.workflows import PackagePublishWorkflow
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    BaseWorkflowData,
    PackagePublishWorkflowData,
    PackagePublishWorkflowDynamicData,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestBase,
)
from debusine.tasks.models import (
    BaseDynamicTaskData,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


@dataclass
class ExpectedCopy:
    """
    A summary of an expected copy.

    This exists to abbreviate tests, and is only useful for copying
    source/binary artifacts.
    """

    artifact: Artifact
    component: str
    section: str
    priority: str | None = None

    @property
    def source_items(self) -> list[str]:
        return [f"{self.artifact.id}@artifacts"]

    @property
    def variables(self) -> dict[str, str]:
        variables = {"component": self.component, "section": self.section}
        if self.priority is not None:
            variables["priority"] = self.priority
        return variables


class PackagePublishWorkflowTests(WorkflowTestBase[PackagePublishWorkflow]):
    """Unit tests for :py:class:`PackagePublishWorkflow`."""

    scenario = scenarios.DefaultContext()
    source_artifact: ClassVar[Artifact]
    binary_artifact: ClassVar[Artifact]

    @classmethod
    def setUpTestData(cls) -> None:
        """Set up common data for tests."""
        super().setUpTestData()
        cls.playground.create_group_role(
            cls.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[cls.scenario.user],
        )
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello"
        )
        cls.binary_artifact = (
            cls.playground.create_minimal_binary_package_artifact("hello")
        )

    def create_suite_collection(
        self, name: str, *, workspace: Workspace | None = None
    ) -> Collection:
        """Create a `debian:suite` collection."""
        return self.playground.create_collection(
            name=name, category=CollectionCategory.SUITE, workspace=workspace
        )

    def create_package_publish_workflow(
        self, task_data: PackagePublishWorkflowData
    ) -> PackagePublishWorkflow:
        """Create a package_publish workflow."""
        wr = self.playground.create_workflow(
            task_name="package_publish", task_data=task_data
        )
        return self.get_workflow(wr)

    def test_create_orchestrator(self) -> None:
        """A PackagePublishWorkflow can be instantiated."""
        source_artifact = self.source_artifact.pk
        binary_artifacts = (self.binary_artifact.pk,)
        target_suite = self.create_suite_collection("bookworm").pk

        workflow = self.create_package_publish_workflow(
            PackagePublishWorkflowData(
                source_artifact=source_artifact,
                binary_artifacts=LookupMultiple(binary_artifacts),
                target_suite=target_suite,
            )
        )

        self.assertEqual(workflow.data.source_artifact, source_artifact)
        self.assertEqual(
            workflow.data.binary_artifacts, LookupMultiple(binary_artifacts)
        )
        self.assertEqual(workflow.data.target_suite, target_suite)

    def orchestrate(
        self,
        task_data: PackagePublishWorkflowData,
        source_artifact: Artifact,
        architectures: Sequence[str],
        *,
        workspace: Workspace | None = None,
    ) -> WorkRequest:
        """Create and orchestrate a PackagePublishWorkflow."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Example pipeline for package publishing."""

            def populate(self) -> None:
                """Populate the pipeline."""
                sbuild = self.work_request_ensure_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(source_artifact=source_artifact.id),
                        target_distribution="debian:sid",
                        architectures=list(architectures) or ["all"],
                    ),
                    workflow_data=WorkRequestWorkflowData(
                        display_name="sbuild", step="sbuild"
                    ),
                )
                self.orchestrate_child(sbuild)

                package_publish = self.work_request_ensure_child_workflow(
                    task_name="package_publish",
                    task_data=task_data,
                    workflow_data=WorkRequestWorkflowData(
                        display_name="publish packages", step="package_publish"
                    ),
                )

                if task_data.source_artifact is not None:
                    self.requires_artifact(
                        package_publish, task_data.source_artifact
                    )
                self.requires_artifact(
                    package_publish, task_data.binary_artifacts
                )
                self.orchestrate_child(package_publish)

        template = self.playground.create_workflow_template(
            name="examplepipeline-template",
            task_name="examplepipeline",
            workspace=workspace,
        )
        root = self.playground.create_workflow(task_name="examplepipeline")
        self.assertEqual(root.workspace, template.workspace)
        self.assertTrue(orchestrate_workflow(root))

        return root

    def assert_copies(
        self,
        work_request: WorkRequest,
        target_collection: Collection,
        expected_copies: Sequence[ExpectedCopy],
    ) -> None:
        """Assert that a task declares the expected copies."""
        self.assertEqual(
            work_request.task_data,
            {
                "copies": [
                    {
                        "source_items": copy.source_items,
                        "target_collection": target_collection.id,
                        "unembargo": False,
                        "replace": False,
                        "variables": copy.variables,
                    }
                    for copy in expected_copies
                ]
            },
        )

    def assert_updates_indexes(
        self, package_publish: WorkRequest, target_suite: Collection
    ) -> None:
        """Assert that the workflow updates indexes after copying."""
        copy_collection_items = package_publish.children.get(
            task_type=TaskTypes.SERVER, task_name="copycollectionitems"
        )
        self.playground.assign_worker(copy_collection_items)
        trigger_update_suites = package_publish.children.get(
            task_type=TaskTypes.INTERNAL,
            task_name="workflow",
            workflow_data_json__step="trigger-update-suites",
        )

        self.assertQuerySetEqual(
            trigger_update_suites.dependencies.all(), [copy_collection_items]
        )

        copy_collection_items_result = run_server_task.apply(
            args=(copy_collection_items.id,)
        )
        self.assertTrue(copy_collection_items_result.get())

        trigger_update_suites.refresh_from_db()
        self.assertEqual(
            trigger_update_suites.status, WorkRequest.Statuses.PENDING
        )

        self.schedule_and_run_workflow_callback(package_publish)
        trigger_update_suites.refresh_from_db()
        self.assertEqual(
            trigger_update_suites.status, WorkRequest.Statuses.COMPLETED
        )
        self.assertEqual(
            trigger_update_suites.result, WorkRequest.Results.SUCCESS
        )

        update_suites = target_suite.workspace.workrequest_set.get(
            task_type=TaskTypes.WORKFLOW, task_name="update_suites"
        )
        self.assertEqual(update_suites.status, WorkRequest.Statuses.PENDING)
        self.assertEqual(
            update_suites.task_data, {"only_suites": [target_suite.name]}
        )

        self.schedule_and_run_workflow(update_suites)
        update_suites.refresh_from_db()
        self.assertEqual(update_suites.status, WorkRequest.Statuses.RUNNING)
        self.assertTrue(update_suites.children.exists())

    @preserve_db_task_registry()
    def test_populate_source_and_binary(self) -> None:
        """Test population with both source and binary artifacts."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", architectures={"any", "all"}
        )
        target_suite = self.create_suite_collection("bookworm")
        architectures = ("amd64", "i386", "all")

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=source_artifact.id,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in architectures
                    ]
                ),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
            ),
            source_artifact=source_artifact,
            architectures=architectures,
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertEqual(package_publish.status, WorkRequest.Statuses.BLOCKED)

        sbuild_workflow = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        binaries, _ = self.simulate_sbuild_workflow_completion(sbuild_workflow)

        package_publish.refresh_from_db()
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(root)

        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        expected = [ExpectedCopy(source_artifact, "main", "misc")]
        for arch in architectures:
            expected += [
                ExpectedCopy(binary_artifact, "main", "misc", "optional")
                for binary_artifact in binaries[arch]
            ]
        self.assert_copies(copy_collection_items, target_suite, expected)
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_source_only(self) -> None:
        """Test population with only a source artifact."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1", architectures={"all"}
        )
        target_suite = self.create_suite_collection("bookworm")

        with self.assertLogs(
            logger="debusine.server.workflows.package_publish",
            level=logging.WARNING,
        ) as log:
            root = self.orchestrate(
                task_data=PackagePublishWorkflowData(
                    source_artifact=source_artifact.id,
                    target_suite=(
                        f"{target_suite.name}@{CollectionCategory.SUITE}"
                    ),
                ),
                source_artifact=source_artifact,
                architectures=[],
            )
        self.assertEqual(
            [record.message for record in log.records],
            [
                f"No component specified when adding hello_1.0-1 to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'main'",
                f"No section specified when adding hello_1.0-1 to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'misc'",
            ],
        )

        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )
        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(source_artifact, "main", "misc")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_source_upload_only(self) -> None:
        """Test population with only a source upload artifact."""
        upload_artifacts = self.playground.create_upload_artifacts(
            src_name="hello",
            src_architectures={"any", "all"},
            binary=False,
            section="editors",
        )
        target_suite = self.create_suite_collection("bookworm")

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
            ),
            source_artifact=upload_artifacts.source,
            architectures=[],
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(upload_artifacts.source, "main", "editors")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_missing_binary_fields(self) -> None:
        """Population warns about missing binary component/section/priority."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1", architectures={"all"}
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                package="hello",
                version="1.0-1",
                architecture="all",
            )
        )
        target_suite = self.create_suite_collection("bookworm")

        with self.assertLogs(
            logger="debusine.server.workflows.package_publish",
            level=logging.WARNING,
        ) as log:
            root = self.orchestrate(
                task_data=PackagePublishWorkflowData(
                    binary_artifacts=LookupMultiple((binary_artifact.id,)),
                    target_suite=(
                        f"{target_suite.name}@{CollectionCategory.SUITE}"
                    ),
                ),
                source_artifact=source_artifact,
                architectures=[],
            )
        self.assertEqual(
            [record.message for record in log.records],
            [
                f"No component specified when adding hello_1.0-1_all to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'main'",
                f"No section specified when adding hello_1.0-1_all to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to 'misc'",
                f"No priority specified when adding hello_1.0-1_all to "
                f"bookworm@{CollectionCategory.SUITE}; defaulting to "
                f"'optional'",
            ],
        )

        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )
        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(binary_artifact, "main", "misc", "optional")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_non_main_component(self) -> None:
        """Population handles sections with non-main components."""
        upload_artifacts = self.playground.create_upload_artifacts(
            src_name="hello",
            src_architectures={"any", "all"},
            binary=False,
            section="contrib/games",
        )
        target_suite = self.create_suite_collection("bookworm")

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
            ),
            source_artifact=upload_artifacts.source,
            architectures=[],
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(upload_artifacts.source, "contrib", "games")],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_honours_suite_variables(self) -> None:
        """Population honours ``suite_variables``."""
        source_artifact = self.playground.create_source_artifact(
            name="hello", version="1.0-1", architectures={"all"}
        )
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello",
                srcpkg_version="1.0-1",
                package="hello",
                version="1.0-1",
                architecture="all",
            )
        )
        target_suite = self.create_suite_collection("bookworm")

        with self.assertNoLogs(
            logger="debusine.server.workflows.package_publish"
        ):
            root = self.orchestrate(
                task_data=PackagePublishWorkflowData(
                    source_artifact=source_artifact.id,
                    binary_artifacts=LookupMultiple((binary_artifact.id,)),
                    target_suite=(
                        f"{target_suite.name}@{CollectionCategory.SUITE}"
                    ),
                    suite_variables={
                        "component": "non-free",
                        "section": "editors",
                        "priority": "extra",
                    },
                ),
                source_artifact=source_artifact,
                architectures=[],
            )

        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )
        self.assert_copies(
            copy_collection_items,
            target_suite,
            [
                ExpectedCopy(artifact, "non-free", "editors", "extra")
                for artifact in (source_artifact, binary_artifact)
            ],
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_no_update_indexes(self) -> None:
        """The workflow can be told not to update indexes."""
        upload_artifacts = self.playground.create_upload_artifacts(
            src_name="hello",
            src_architectures={"any", "all"},
            binary=False,
            section="editors",
        )
        target_suite = self.create_suite_collection("bookworm")

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                source_artifact=upload_artifacts.upload.id,
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                update_indexes=False,
            ),
            source_artifact=upload_artifacts.source,
            architectures=[],
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        self.assert_copies(
            copy_collection_items,
            target_suite,
            [ExpectedCopy(upload_artifacts.source, "main", "editors")],
        )
        self.assertFalse(
            package_publish.children.filter(
                task_type=TaskTypes.INTERNAL,
                task_name="workflow",
                workflow_data_json__step="trigger-update-suites",
            ).exists()
        )

    @preserve_db_task_registry()
    def test_populate_package_build_logs(self) -> None:
        """Test population with binary artifacts and copying build logs."""
        source_workspace = self.playground.create_workspace(name="source")
        self.playground.create_group_role(
            source_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        target_workspace = self.playground.create_workspace(name="target")
        self.playground.create_group_role(
            target_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE, workspace=target_workspace
        )
        source_workspace.set_inheritance([target_workspace])
        target_suite = self.create_suite_collection(
            "bookworm", workspace=target_workspace
        )
        source_build_logs_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.PACKAGE_BUILD_LOGS,
                workspace=source_workspace,
            )
        )
        target_build_logs_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.PACKAGE_BUILD_LOGS,
                workspace=target_workspace,
            )
        )
        source_artifact = self.playground.create_source_artifact(
            name="hello", architectures={"any", "all"}
        )
        architectures = ("amd64", "i386", "all")
        binary_artifacts_lookup = [
            f"internal@collections/name:build-{arch}" for arch in architectures
        ]

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple.model_validate(
                    binary_artifacts_lookup
                ),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                unembargo=True,
            ),
            source_artifact=source_artifact,
            architectures=architectures,
            workspace=source_workspace,
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertEqual(package_publish.status, WorkRequest.Statuses.BLOCKED)

        sbuild_workflow = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        binaries, _ = self.simulate_sbuild_workflow_completion(sbuild_workflow)

        package_publish.refresh_from_db()
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(root)

        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        expected_copies: list[dict[str, Any]] = []
        for arch in architectures:
            for binary_artifact in binaries[arch]:
                expected_copies.append(
                    {
                        "source_items": [f"{binary_artifact.id}@artifacts"],
                        "target_collection": target_suite.id,
                        "unembargo": True,
                        "replace": False,
                        "variables": {
                            "component": "main",
                            "section": "misc",
                            "priority": "optional",
                        },
                    }
                )
        expected_copies.append(
            {
                "source_items": [
                    {
                        "collection": source_build_logs_collection.id,
                        "lookup_filters": [
                            [
                                "same_work_request",
                                binary_artifacts_lookup,
                            ]
                        ],
                    }
                ],
                "target_collection": target_build_logs_collection.id,
                "unembargo": True,
                "replace": False,
            }
        )
        self.assertEqual(
            copy_collection_items.task_data, {"copies": expected_copies}
        )
        self.assert_updates_indexes(package_publish, target_suite)

    @preserve_db_task_registry()
    def test_populate_package_task_history(self) -> None:
        """Test population with binary artifacts and copying build logs."""
        source_workspace = self.playground.create_workspace(name="source")
        self.playground.create_group_role(
            source_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        target_workspace = self.playground.create_workspace(name="target")
        self.playground.create_group_role(
            target_workspace, Workspace.Roles.OWNER, users=[self.scenario.user]
        )
        self.playground.create_singleton_collection(
            CollectionCategory.ARCHIVE, workspace=target_workspace
        )
        source_workspace.set_inheritance([target_workspace])
        target_suite = self.create_suite_collection(
            "bookworm", workspace=target_workspace
        )
        source_task_history_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.TASK_HISTORY,
                workspace=source_workspace,
            )
        )
        target_task_history_collection = (
            self.playground.create_singleton_collection(
                CollectionCategory.TASK_HISTORY,
                workspace=target_workspace,
            )
        )
        source_artifact = self.playground.create_source_artifact(
            name="hello", architectures={"any", "all"}
        )
        architectures = ("amd64", "i386", "all")
        binary_artifacts_lookup = [
            f"internal@collections/name:build-{arch}" for arch in architectures
        ]

        root = self.orchestrate(
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple.model_validate(
                    binary_artifacts_lookup
                ),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                unembargo=True,
            ),
            source_artifact=source_artifact,
            architectures=architectures,
            workspace=source_workspace,
        )
        package_publish = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="package_publish"
        )
        self.assertEqual(package_publish.status, WorkRequest.Statuses.BLOCKED)

        sbuild_workflow = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        binaries, _ = self.simulate_sbuild_workflow_completion(sbuild_workflow)

        package_publish.refresh_from_db()
        self.assertEqual(package_publish.status, WorkRequest.Statuses.PENDING)
        self.schedule_and_run_workflow(root)

        copy_collection_items = package_publish.children.get(
            task_name="copycollectionitems"
        )

        expected_copies: list[dict[str, Any]] = []
        for arch in architectures:
            for binary_artifact in binaries[arch]:
                expected_copies.append(
                    {
                        "source_items": [f"{binary_artifact.id}@artifacts"],
                        "target_collection": target_suite.id,
                        "unembargo": True,
                        "replace": False,
                        "variables": {
                            "component": "main",
                            "section": "misc",
                            "priority": "optional",
                        },
                    }
                )
        expected_copies.append(
            {
                "source_items": [
                    {
                        "collection": source_task_history_collection.id,
                        "lookup_filters": [
                            ["same_workflow", binary_artifacts_lookup]
                        ],
                    }
                ],
                "target_collection": target_task_history_collection.id,
                "unembargo": True,
                "replace": False,
            }
        )
        self.assertEqual(
            copy_collection_items.task_data, {"copies": expected_copies}
        )
        self.assert_updates_indexes(package_publish, target_suite)

    def test_compute_system_required_tags(self) -> None:
        upload_artifact = (
            self.playground.create_upload_artifacts(src_name="hello")
        ).upload
        target_suite = self.create_suite_collection("bookworm")

        workflow = self.create_package_publish_workflow(
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple((upload_artifact.id,)),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                source_artifact=self.source_artifact.id,
            )
        )

        self.assertCountEqual(
            workflow.compute_system_required_tags(),
            [wtags.WORKER_TYPE_NOT_ASSIGNABLE],
        )

    def test_compute_dynamic_data_binary_package(self) -> None:
        upload_artifact = (
            self.playground.create_upload_artifacts(src_name="hello")
        ).upload
        target_suite = self.create_suite_collection("bookworm")

        workflow = self.create_package_publish_workflow(
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple((upload_artifact.id,)),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                source_artifact=self.source_artifact.id,
            )
        )

        self.assertEqual(
            workflow.compute_dynamic_data(TaskDatabase(workflow.work_request)),
            PackagePublishWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello → "
                f"bookworm@{CollectionCategory.SUITE}",
                source_artifact_id=self.source_artifact.id,
                target_suite_id=target_suite.id,
                binary_artifacts_ids=[upload_artifact.id],
            ),
        )

    def test_get_input_artifacts(self) -> None:
        upload_artifact = (
            self.playground.create_upload_artifacts(src_name="hello")
        ).upload
        target_suite = self.create_suite_collection("bookworm")

        workflow = self.create_package_publish_workflow(
            PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple((upload_artifact.id,)),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
                source_artifact=self.source_artifact.id,
            ),
        )

        workflow.dynamic_data = workflow.build_dynamic_data(
            TaskDatabase(workflow.work_request)
        )

        assert workflow.data.source_artifact is not None

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[upload_artifact.id],
                ),
            ],
        )

    def test_get_input_artifacts_without_source_artifact(self) -> None:
        upload_artifact = (
            self.playground.create_upload_artifacts(src_name="hello")
        ).upload
        target_suite = self.create_suite_collection("bookworm")

        workflow = self.create_package_publish_workflow(
            task_data=PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple((upload_artifact.id,)),
                target_suite=(
                    f"{target_suite.name}@{CollectionCategory.SUITE}"
                ),
            ),
        )

        workflow.dynamic_data = workflow.build_dynamic_data(
            TaskDatabase(workflow.work_request)
        )

        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactMultiple(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[upload_artifact.id],
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the client_utils.py methods."""

import hashlib
import logging
from textwrap import dedent
from unittest import mock

import requests
import responses
from debian.deb822 import Changes, Deb822

from debusine.artifacts.playground import ArtifactPlayground
from debusine.client.client_utils import (
    copy_file,
    dget,
    download_file,
    get_debian_package,
    get_url_contents_sha256sum,
    requests_put_or_connection_error,
)
from debusine.client.exceptions import (
    ClientConnectionError,
    ContentTooLargeError,
    ContentValidationError,
)
from debusine.test import TestCase


class RequestsPutOrConnectionErrorTests(TestCase):
    """Tests for requests_put_or_connection_error."""

    def test_requests_put_raise_exc_without_request(self) -> None:
        """
        requests_put_or_connection_error handles exceptions without a request.

        This is for 100% branch coverage; as far as we know,
        :py:meth:`requests.put` always sets the ``request`` attribute on
        exceptions.
        """
        patcher = mock.patch("debusine.client.client_utils.requests.put")
        mocked = patcher.start()
        mocked.side_effect = requests.exceptions.RequestException(request=None)
        self.addCleanup(patcher.stop)

        with self.assertRaisesRegex(
            ClientConnectionError, r"^Cannot connect\. Error: "
        ):
            requests_put_or_connection_error()


class DgetDownloadTests(TestCase):
    """Tests for our dget client."""

    def setUp(self) -> None:
        """Configure fake http responses."""
        super().setUp()
        self.workdir = self.create_temporary_directory()

        self.r_mock = responses.RequestsMock(assert_all_requests_are_fired=True)
        self.r_mock.start()
        self.addCleanup(self.r_mock.stop)
        self.addCleanup(self.r_mock.reset)

        # Swallow log messages
        log = logging.getLogger("debusine.client.client_utils")
        propagate_setting = log.propagate
        log.propagate = False
        self.addCleanup(lambda: setattr(log, "propagate", propagate_setting))

        self.bodies = {
            "foo.deb": b"abc123",
            "foo_1.0.orig.tar.xz": b"test data",
            "foo_1.0-2.debian.tar.xz": b"more test data",
            "foo_1.0-2_source.buildinfo": b"Format: 1.0\nSource: foo\n",
        }
        for filename, body in self.bodies.items():
            self.r_mock.get(f"http://example.com/{filename}", body=body)
        self.set_dsc_response()
        self.set_changes_response()
        self.expected_stats = {
            "foo.deb": {
                "sha256": "6ca13d52ca70c883e0f0bb101e425a89e8624de51db2d239259"
                "3af6a84118090",
                "sha1": "6367c48dd193d56ea7b0baad25b19455e529f5ee",
                "md5": "e99a18c428cb38d5f260853678922e03",
                "size": 6,
            },
        }

    def _set_response(
        self,
        filename: str,
        body: str,
        files: list[str],
        break_hashes: set[str] = set(),
        break_sizes: set[str] = set(),
    ) -> None:
        file_type: type[Deb822]
        if filename.endswith(".changes"):
            file_type = Changes
        else:
            file_type = Deb822
        full_body = (
            body
            + ArtifactPlayground.hash_deb822_files(
                file_type=file_type,
                file_contents={file: self.bodies[file] for file in files},
                break_hashes=break_hashes,
                break_sizes=break_sizes,
            )
        ).encode('utf-8')
        self.bodies[filename] = full_body
        self.r_mock.upsert(
            responses.GET, f"http://example.com/{filename}", body=full_body
        )

    def set_dsc_response(
        self,
        break_hashes: set[str] = set(),
        break_sizes: set[str] = set(),
    ) -> None:
        """Add / replace the .dsc response."""
        self._set_response(
            "foo_1.0-2.dsc",
            dedent(
                """
                Format: 3.0 (quilt)
                Source: foo
                Package-List:
                 foo deb unknown optional arch=any
                """
            ).lstrip(),
            ["foo_1.0.orig.tar.xz", "foo_1.0-2.debian.tar.xz"],
            break_hashes=break_hashes,
            break_sizes=break_sizes,
        )

    def set_changes_response(self) -> None:
        """Add / replace the .changes response."""
        self._set_response(
            "foo_1.0-2_source.changes",
            dedent(
                """
                Format: 1.8
                Source: foo
                """
            ).lstrip(),
            [
                "foo_1.0-2.dsc",
                "foo_1.0-2.debian.tar.xz",
                "foo_1.0-2_source.buildinfo",
            ],
        )

    def test_copy_file_stats(self) -> None:
        src = self.workdir / "src.deb"
        dest = self.workdir / "foo.deb"
        src.write_bytes(self.bodies["foo.deb"])
        stats = copy_file(src, dest)
        self.assertEqual(self.expected_stats["foo.deb"], stats)

    def test_download_file_downloads(self) -> None:
        """Check `download_file` writes the expected file."""
        dest = self.workdir / "foo.deb"
        download_file("http://example.com/foo.deb", dest)
        self.assertTrue(dest.exists())
        self.assertEqual(dest.read_bytes(), b"abc123")

    def test_download_file_stats(self) -> None:
        """Check the return value of `download_file`."""
        dest = self.workdir / "foo.deb"
        stats = download_file("http://example.com/foo.deb", dest)
        self.assertEqual(self.expected_stats["foo.deb"], stats)

    def test_download_file_logging(self) -> None:
        """Ensure `download_file` logs its requests."""
        dest = self.workdir / "foo.deb"
        with self.assertLogsContains(
            message="Downloading http://example.com/foo.deb...",
            logger="debusine.client.client_utils",
        ):
            download_file("http://example.com/foo.deb", dest)

    def test_download_file_size(self) -> None:
        """Ensure `download_file` logs the size of files it's downloading."""
        dest = self.workdir / "big.deb"
        self.r_mock.get(
            "http://example.com/big.deb",
            body=b"1234567890" * 1000,
            headers={
                "Content-Length": "10000",
            },
        )
        with self.assertLogs(logger="debusine.client.client_utils") as cm:
            download_file("http://example.com/big.deb", dest)
        self.assertEqual(2, len(cm.records))
        messages = [record.message for record in cm.records]
        self.assertIn("Size: 0.01 MiB", messages)

    def test_dget_file(self) -> None:
        """Test that `dget` can download a .deb."""
        dget("http://example.com/foo.deb", self.workdir)
        dest = self.workdir / "foo.deb"
        self.assertTrue(dest.exists())

    def test_dget_subdir(self) -> None:
        """Ensure `dget` correctly handles sub-directories."""
        dest = self.workdir / "test.deb"
        self.r_mock.get(
            "http://example.com/subdir/test.deb",
            body=b"abc123",
        )
        dget("http://example.com/subdir/test.deb", self.workdir)
        self.assertTrue(dest.exists())

    def test_dget_dsc(self) -> None:
        """Test that `dget` can download a .dsc."""
        dget("http://example.com/foo_1.0-2.dsc", self.workdir)
        self.assertTrue((self.workdir / "foo_1.0-2.dsc").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.debian.tar.xz").exists())
        self.assertTrue((self.workdir / "foo_1.0.orig.tar.xz").exists())

    def test_dget_source_changes(self) -> None:
        """Test that `dget` can download a _source.changes."""
        dget("http://example.com/foo_1.0-2_source.changes", self.workdir)
        self.assertTrue((self.workdir / "foo_1.0-2_source.changes").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.dsc").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.debian.tar.xz").exists())
        # Test that we recurse through the .dsc to find the orig tarball
        self.assertTrue((self.workdir / "foo_1.0.orig.tar.xz").exists())

    def test_dget_only_downloads_files_once(self) -> None:
        """Check that `dget` downloads aren't unnecessarily repeated."""
        with self.assertLogs(logger="debusine.client.client_utils") as cm:
            dget("http://example.com/foo_1.0-2_source.changes", self.workdir)
        records = [
            record.message
            for record in cm.records
            if record.message.startswith("Downloading")
        ]
        self.assertEqual(len(set(records)), len(records))
        self.assertTrue((self.workdir / "foo_1.0-2_source.changes").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.dsc").exists())
        self.assertTrue((self.workdir / "foo_1.0-2.debian.tar.xz").exists())
        self.assertTrue((self.workdir / "foo_1.0.orig.tar.xz").exists())

    def test_dget_garbage_dsc(self) -> None:
        """If we can't parse it, `dget` simply downloads the .dsc."""
        self.r_mock.get(
            "http://example.com/garbage.dsc",
            body=b"\0:" * 100,
        )
        dget("http://example.com/garbage.dsc", self.workdir)
        self.assertTrue((self.workdir / "garbage.dsc").exists())

    def test_dget_garbage_changes(self) -> None:
        """If we can't parse it, `dget` simply downloads the .changes."""
        self.r_mock.get(
            "http://example.com/garbage.changes",
            body=b"\0:" * 100,
        )
        dget("http://example.com/garbage.changes", self.workdir)
        self.assertTrue((self.workdir / "garbage.changes").exists())

    def test_dget_verifies_sha256(self) -> None:
        """`dget` verifies SHA256 hashes."""
        self.set_dsc_response(break_hashes={"sha256"})
        with self.assertRaisesRegex(
            ContentValidationError,
            r"Downloaded file has mis-matching sha256 for "
            r"foo_1\.0\.orig\.tar\.xz \([0-9a-f]{64} != [0-9a-f]{64}\)",
        ):
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)

    def test_dget_verifies_sha1(self) -> None:
        """`dget` verifies SHA1 hashes."""
        self.set_dsc_response(break_hashes={"sha1"})
        with self.assertRaisesRegex(
            ContentValidationError,
            r"Downloaded file has mis-matching sha1 for "
            r"foo_1\.0\.orig\.tar\.xz \([0-9a-f]{40} != [0-9a-f]{40}\)",
        ):
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)

    def test_dget_verifies_md5(self) -> None:
        """`dget` verifies MD5 hashes."""
        self.set_dsc_response(break_hashes={"md5"})
        with self.assertRaisesRegex(
            ContentValidationError,
            r"Downloaded file has mis-matching md5 for "
            r"foo_1\.0\.orig\.tar\.xz \([0-9a-f]{32} != [0-9a-f]{32}\)",
        ):
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)

    def test_dget_verifies_size(self) -> None:
        """`dget` will spot size mismatch between entries in dsc file."""
        self.set_dsc_response(break_sizes={"sha256"})
        with self.assertRaises(ContentValidationError) as cm:
            dget("http://example.com/foo_1.0-2.dsc", self.workdir)
        self.assertEqual(
            "foo_1.0-2.dsc has mis-matching size for foo_1.0.orig.tar.xz "
            "(9 != 10)",
            str(cm.exception),
        )

    def test_dget_verifies_previously_downloaded_hash(self) -> None:
        """`dget` spots files differing between .dsc and .changes files."""
        self.set_dsc_response(break_hashes={"sha256"})
        self.set_changes_response()
        with self.assertRaisesRegex(
            ContentValidationError,
            r"foo_1.0-2\.dsc has mis-matching sha256 for "
            r"foo_1\.0-2\.debian\.tar\.xz \([0-9a-f]{64} != [0-9a-f]{64}\)",
        ):
            dget("http://example.com/foo_1.0-2_source.changes", self.workdir)

    def test_dget_rejects_path_traversal(self) -> None:
        """`dget` will refuse to download a dsc file with / in filenames."""
        self.r_mock.get(
            "http://example.com/traversal.dsc",
            body=dedent(
                """
                Format: 3.0 (quilt)
                Source: foo
                Package-List:
                 foo deb unknown optional arch=any
                Checksums-Sha1:
                 f48dd853820860816c75d54d0f584dc863327a7c 9 foo/bar.tar.xz
                Files:
                 eb733a00c0c9d336e65691a37ab54293 9 foo/bar.tar.xz
                """
            )
            .lstrip()
            .encode('utf-8'),
        )
        with self.assertRaises(ContentValidationError) as cm:
            dget("http://example.com/traversal.dsc", self.workdir)
        self.assertEqual(
            "traversal.dsc contains invalid file name foo/bar.tar.xz",
            str(cm.exception),
        )

    def test_get_debian_package_local_file(self) -> None:
        """`get_debian_package` will return a Path to a local .deb file."""
        source = self.workdir / "foo.deb"
        source.write_bytes(b"test file")
        with get_debian_package(str(source)) as package:
            self.assertEqual(package, source)

    def test_get_debian_package_file_url(self) -> None:
        """`get_debian_package` will return a Path to a file://...deb URL."""
        source = self.workdir / "foo.deb"
        source.write_bytes(b"test file")
        with get_debian_package(f"file://{source}") as package:
            self.assertEqual(package, source)

    def test_get_debian_package_http_url(self) -> None:
        """`get_debian_package` will download a deb by http."""
        with get_debian_package("http://example.com/foo.deb") as package:
            self.assertEqual(package.read_bytes(), b"abc123")

    def test_get_debian_package_ftp_url(self) -> None:
        """`get_debian_package` can't take ftp:// URLs."""
        with (
            self.assertRaisesRegex(ValueError, r"Not a supported URL scheme:"),
            get_debian_package("ftp://example.com/foo.deb"),
        ):
            pass  # pragma: no cover


class GetUrlContentsSha256sumTests(TestCase):
    """Tests for get_url_contents_sha256sum function."""

    @responses.activate
    def test_get_url_contents_sha256sum(self) -> None:
        """get_url_contents_sha256sum function return contents and sha256."""
        url = "https://example.com"
        max_size = 100
        content = b"Test content"
        sha256sum = hashlib.sha256(content).hexdigest()

        responses.add(responses.GET, url, body=content)

        returned_content, returned_sha256sum = get_url_contents_sha256sum(
            url, max_size
        )

        self.assertEqual(returned_content, content)
        self.assertEqual(returned_sha256sum, sha256sum)

    @responses.activate
    def test_get_url_contents_sha256sum_too_large(self) -> None:
        """get_url_contents_sha256sum function raise ContentTooLargeError."""
        url = "https://example.com"
        max_size = 10
        content = b'Test content that is too large for the specified max_size'

        responses.add(responses.GET, url, body=content)

        message = "Content size exceeds maximum allowed size of 10 bytes."
        with self.assertRaisesRegex(ContentTooLargeError, message):
            get_url_contents_sha256sum(url, max_size)

    def test_file_url_without_allow_file(self) -> None:
        """Without allow_file=True, file:// URLs are rejected."""
        path = self.create_temporary_file()

        with self.assertRaises(requests.exceptions.InvalidSchema):
            get_url_contents_sha256sum(f"file://{path}", 10)

    def test_file_url_with_allow_file(self) -> None:
        """With allow_file=True, file:// URLs work."""
        max_size = 100
        content = b"Test content"
        sha256sum = hashlib.sha256(content).hexdigest()
        path = self.create_temporary_file(contents=content)

        returned_content, returned_sha256sum = get_url_contents_sha256sum(
            f"file://{path}", max_size, allow_file=True
        )

        self.assertEqual(returned_content, content)
        self.assertEqual(returned_sha256sum, sha256sum)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, collection management commands."""

import abc
import argparse
from collections.abc import Iterable
from typing import override

import rich
import rich.json
from rich.table import Table

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.base import (
    ModelCommand,
    OptionalInputDataCommand,
    RequiredInputDataCommand,
    WorkspaceCommand,
)
from debusine.client.models import CollectionData, CollectionDataNew


class CollectionCommand(
    ModelCommand[CollectionData], WorkspaceCommand, abc.ABC
):
    """Common infrastructure for collection commands."""

    def markup_category(self, collection: CollectionData) -> str:
        """Markup a collection category for rich."""
        if collection.category == CollectionCategory.WORKFLOW_INTERNAL:
            return collection.category
        category_url = self.debusine.webui_url(
            collection.get_link_url("webui_category")
        )
        return f"[link={category_url}]{collection.category}[/]"

    def _url(self, instance: CollectionData) -> str:
        return self.debusine.webui_url(instance.get_link_url("webui_self"))

    def _list_rich(self, collections: Iterable[CollectionData]) -> None:
        """Print the list as a table."""
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("ID")
        table.add_column("Name")
        table.add_column("Category")
        table.add_column("Retention (history)")
        table.add_column("Retention (metadata)")
        table.add_column("Data")
        for collection in collections:
            url = self._url(collection)
            table.add_row(
                f"{collection.id}",
                f"[link={url}]{collection.name}[/]",
                self.markup_category(collection),
                (
                    str(period)
                    if (period := collection.full_history_retention_period)
                    is not None
                    else "unlimited"
                ),
                (
                    str(period)
                    if (period := collection.metadata_only_retention_period)
                    is not None
                    else "unlimited"
                ),
                (
                    rich.json.JSON.from_data(collection.data)
                    if collection.data
                    else None
                ),
            )
        rich.print(table)

    def _show_rich(self, collection: CollectionData) -> None:
        """Show a collection for humans."""
        url = self._url(collection)
        table = Table(box=rich.box.SIMPLE, show_header=False)
        table.add_column("Name", justify="right")
        table.add_row("ID:", f"#{collection.id}")
        table.add_row("Name:", f"[link={url}]{collection.name}[/]")
        table.add_row("Category:", self.markup_category(collection))
        full = (
            f"{period} days"
            if (period := collection.full_history_retention_period) is not None
            else "unlimited"
        )
        meta = (
            f"{period} days"
            if (period := collection.metadata_only_retention_period) is not None
            else "unlimited"
        )
        table.add_row("Full history retention:", full)
        table.add_row("Metadata only retention:", meta)
        table.add_row("Data:", rich.json.JSON.from_data(collection.data))
        if collection.stats:
            table.add_row("Stats:", rich.json.JSON.from_data(collection.stats))
        rich.print(table)


class List(CollectionCommand, group="collection"):
    """List collections in a workspace."""

    def run(self) -> None:
        """Run the command."""
        with self._api_call_or_fail():
            collections = self.debusine.collection_iter(self.workspace)
            self.list(collections)


class SingleCollectionCommand(WorkspaceCommand, abc.ABC):
    """Command operating on a single collection."""

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Identify the selected collection."""
        super().__init__(parser, args)
        if "@" in (cat := args.category_or_fullname):
            self.collection_name, self.collection_category = cat.split("@", 1)
            if args.name is not None:
                parser.error(
                    "collection name is given"
                    f" both in `{args.category_or_fullname}`"
                    " and in the name argument"
                )
        else:
            self.collection_category = cat
            self.collection_name = args.name or "_"

    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        """Configure the ArgumentParser for this subcommand."""
        super().configure(parser)
        parser.add_argument(
            "category_or_fullname",
            type=str,
            metavar="[name@]category",
            help="collection category or name@category full name",
        )
        parser.add_argument(
            "name",
            type=str,
            nargs="?",
            help="collection name, only needed if not specified"
            " in the first argument. Defaults to '_'.",
        )


class Show(SingleCollectionCommand, CollectionCommand, group="collection"):
    """Show an existing collection."""

    def run(self) -> None:
        """Run the command."""
        with self._api_call_or_fail():
            collection = self.debusine.collection_get(
                self.workspace,
                self.collection_category,
                self.collection_name,
                with_stats=True,
            )
        self.show(collection)


def retention_period(value: str) -> int | str:
    """Validate a retention period from the command line."""
    if value.isdigit():
        return int(value)
    elif value == "-":
        return "-"
    else:
        raise argparse.ArgumentError(
            argument=None, message="retention period must be a number or '-'"
        )


class CollectionInput(CollectionCommand, abc.ABC):
    """Common code for commands that input data for collections."""

    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        """Configure the ArgumentParser for this subcommand."""
        super().configure(parser)
        parser.add_argument(
            "--full-history-retention-period",
            type=retention_period,
            help="set full history retention period (in days)."
            " Use a dash ('-') to unset",
        )
        parser.add_argument(
            "--metadata-only-retention-period",
            type=retention_period,
            help="set metadata only retention period (in days)."
            " Use a dash ('-') to unset",
        )

    def update_collection(
        self, collection: CollectionDataNew | CollectionData
    ) -> None:
        """Update a collection record with data from command line arguments."""
        if (period := self.args.full_history_retention_period) is not None:
            if period == "-":
                collection.full_history_retention_period = None
            else:
                collection.full_history_retention_period = int(period)

        if (period := self.args.metadata_only_retention_period) is not None:
            if period == "-":
                collection.metadata_only_retention_period = None
            else:
                collection.metadata_only_retention_period = int(period)


class Create(
    SingleCollectionCommand,
    RequiredInputDataCommand,
    CollectionInput,
    group="collection",
):
    """Create a new collection."""

    def update_collection(
        self, collection: CollectionDataNew | CollectionData
    ) -> None:
        """Update a collection record with data from command line arguments."""
        super().update_collection(collection)
        collection.data = self.input_data

    def run(self) -> None:
        """Run the command."""
        collection = CollectionDataNew(
            category=self.collection_category,
            name=self.collection_name,
            data={},
        )
        self.update_collection(collection)

        with self._api_call_or_fail():
            collection_new = self.debusine.collection_create(
                self.workspace, collection
            )

        self.show(collection_new)


class Manage(
    SingleCollectionCommand,
    OptionalInputDataCommand,
    CollectionInput,
    group="collection",
):
    """Configure an existing collection."""

    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        """Configure the ArgumentParser for this subcommand."""
        super().configure(parser)
        parser.add_argument("--rename", type=str, help="rename collection")

    def update_collection(
        self, collection: CollectionDataNew | CollectionData
    ) -> None:
        """Update a collection record with data from command line arguments."""
        super().update_collection(collection)
        if self.input_data is not None:
            collection.data = self.input_data

    def run(self) -> None:
        """Run the command."""
        with self._api_call_or_fail():
            collection = self.debusine.collection_get(
                self.workspace, self.collection_category, self.collection_name
            )
        self.update_collection(collection)

        if (name := self.args.rename) is not None:
            collection.name = name

        with self._api_call_or_fail():
            collection = self.debusine.collection_update(
                self.workspace, collection
            )

        self.show(collection)


class Delete(SingleCollectionCommand, WorkspaceCommand, group="collection"):
    """Delete a collection."""

    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        """Configure the ArgumentParser for this subcommand."""
        super().configure(parser)
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="force deleting non-empty collections",
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            collection = self.debusine.collection_get(
                self.workspace,
                self.collection_category,
                self.collection_name,
                with_stats=not self.args.force,
            )

        if not self.args.force:
            assert collection.stats is not None
            items: list[str] = []
            for s in collection.stats:
                state = "active" if s["active"] else "deleted"
                items.append(
                    f"{state} {s["category"]} {s["type"]}: {s["count"]}"
                )
            if items:
                self._fail(
                    f"Collection is not empty ({", ".join(items)}):"
                    " use --force to delete it."
                )

        with self._api_call_or_fail():
            self.debusine.collection_delete(self.workspace, collection)

        self.feedback("Collection deleted.")
